import os
from unittest import TestCase

from clayrs.content_analyzer.raw_information_source import SQLDatabase, CSVFile, JSONFile, DATFile
from test import dir_test_files

json_file = os.path.join(dir_test_files, "movies_info_reduced.json")
csv_w_header = os.path.join(dir_test_files, 'movies_info_reduced.csv')
csv_no_header = os.path.join(dir_test_files, 'test_ratings', 'ratings_1591277020.csv')
dat_file = os.path.join(dir_test_files, 'users_70.dat')
tsv_file = os.path.join(dir_test_files, 'random_tsv.tsv')


class TestSQLDatabase(TestCase):

    def test_iter(self):
        self.skipTest("FIX TEST")
        sql = SQLDatabase('localhost', 'root', 'password', 'prova', 'tabella')
        my_iter = iter(sql)
        d1 = {'campo1': 'Francesco', 'campo2': 'Benedetti', 'campo3': 'Polignano'}
        d2 = {'campo1': 'Mario', 'campo2': 'Rossi', 'campo3': 'Roma'}
        d3 = {'campo1': 'Gigio', 'campo2': 'Donnarumma', 'campo3': 'Milano'}
        self.assertDictEqual(next(my_iter), d1)
        self.assertDictEqual(next(my_iter), d2)
        self.assertDictEqual(next(my_iter), d3)


class TestCSVFile(TestCase):

    def test_iter(self):
        csv = CSVFile(csv_w_header)
        my_iter = iter(csv)
        d1 = {"Title": "Jumanji", "Year": "1995", "Rated": "PG", "Released": "15 Dec 1995", "Runtime": "104 min",
              "Genre": "Adventure, Family, Fantasy", "Director": "Joe Johnston",
              "Writer": "Jonathan Hensleigh (screenplay by), Greg Taylor (screenplay by), Jim Strain (screenplay by), Greg Taylor (screen story by), Jim Strain (screen story by), Chris Van Allsburg (screen story by), Chris Van Allsburg (based on the book by)",
              "Actors": "Robin Williams, Jonathan Hyde, Kirsten Dunst, Bradley Pierce",
              "Plot": "After being trapped in a jungle board game for 26 years, a Man-Child wins his release from the game. But, no sooner has he arrived that he is forced to play again, and this time sets the creatures of the jungle loose on the city. Now it is up to him to stop them.",
              "Language": "English, French", "Country": "USA", "Awards": "4 wins & 9 nominations.",
              "Poster": "https://m.media-amazon.com/images/M/MV5BZTk2ZmUwYmEtNTcwZS00YmMyLWFkYjMtNTRmZDA3YWExMjc2XkEyXkFqcGdeQXVyMTQxNzMzNDI@._V1_SX300.jpg",
              "Metascore": "39", "imdbRating": "6.9", "imdbVotes": "260,909", "imdbID": "tt0113497", "Type": "movie",
              "DVD": "25 Jan 2000", "BoxOffice": "N/A", "Production": "Sony Pictures Home Entertainment",
              "Website": "N/A",
              "Response": "True"}
        d2 = {"Title": "Grumpier Old Men", "Year": "1995", "Rated": "PG-13", "Released": "22 Dec 1995",
              "Runtime": "101 min",
              "Genre": "Comedy, Romance", "Director": "Howard Deutch",
              "Writer": "Mark Steven Johnson (characters), Mark Steven Johnson",
              "Actors": "Walter Matthau, Jack Lemmon, Sophia Loren, Ann-Margret",
              "Plot": "Things don't seem to change much in Wabasha County: Max and John are still fighting after 35 years, Grandpa still drinks, smokes, and chases women , and nobody's been able to catch the fabled \"Catfish Hunter\", a gigantic catfish that actually smiles at fishermen who try to snare it. Six months ago John married the new girl in town (Ariel), and people begin to suspect that Max might be missing something similar in his life. The only joy Max claims is left in his life is fishing, but that might change with the new owner of the bait shop.",
              "Language": "English, Italian, German", "Country": "USA", "Awards": "2 wins & 2 nominations.",
              "Poster": "https://m.media-amazon.com/images/M/MV5BMjQxM2YyNjMtZjUxYy00OGYyLTg0MmQtNGE2YzNjYmUyZTY1XkEyXkFqcGdeQXVyMTQxNzMzNDI@._V1_SX300.jpg",
              "Metascore": "46", "imdbRating": "6.6", "imdbVotes": "21,823", "imdbID": "tt0113228", "Type": "movie",
              "DVD": "18 Nov 1997", "BoxOffice": "N/A", "Production": "Warner Home Video", "Website": "N/A",
              "Response": "True"}
        d3 = {"Title": "Toy Story", "Year": "1995", "Rated": "G", "Released": "22 Nov 1995", "Runtime": "81 min",
              "Genre": "Animation, Adventure, Comedy, Family, Fantasy", "Director": "John Lasseter",
              "Writer": "John Lasseter (original story by), Pete Docter (original story by), Andrew Stanton (original story by), Joe Ranft (original story by), Joss Whedon (screenplay by), Andrew Stanton (screenplay by), Joel Cohen (screenplay by), Alec Sokolow (screenplay by)",
              "Actors": "Tom Hanks, Tim Allen, Don Rickles, Jim Varney",
              "Plot": "A little boy named Andy loves to be in his room, playing with his toys, especially his doll named \"Woody\". But, what do the toys do when Andy is not with them, they come to life. Woody believes that he has life (as a toy) good. However, he must worry about Andy's family moving, and what Woody does not know is about Andy's birthday party. Woody does not realize that Andy's mother gave him an action figure known as Buzz Lightyear, who does not believe that he is a toy, and quickly becomes Andy's new favorite toy. Woody, who is now consumed with jealousy, tries to get rid of Buzz. Then, both Woody and Buzz are now lost. They must find a way to get back to Andy before he moves without them, but they will have to pass through a ruthless toy killer, Sid Phillips.",
              "Language": "English", "Country": "USA",
              "Awards": "Nominated for 3 Oscars. Another 23 wins & 17 nominations.",
              "Poster": "https://m.media-amazon.com/images/M/MV5BMDU2ZWJlMjktMTRhMy00ZTA5LWEzNDgtYmNmZTEwZTViZWJkXkEyXkFqcGdeQXVyNDQ2OTk4MzI@._V1_SX300.jpg",
              "Metascore": "95", "imdbRating": "8.3", "imdbVotes": "761,649", "imdbID": "tt0114709", "Type": "movie",
              "DVD": "20 Mar 2001", "BoxOffice": "N/A", "Production": "Buena Vista",
              "Website": "http://www.disney.com/ToyStory", "Response": "True"}

        self.assertDictEqual(next(my_iter), d1)
        self.assertDictEqual(next(my_iter), d2)
        self.assertDictEqual(next(my_iter), d3)

    def test__len_w_header(self):
        csv = CSVFile(csv_w_header)

        self.assertEqual(3, len(csv))

    def test_iter_no_header(self):
        csv = CSVFile(csv_no_header, has_header=False)

        expected_row_1 = {'0': '01', '1': 'a', '2': '0.2333333333333333', '3': '1234567', '4': 'not so good',
                          '5': 'I expected more from this product', '6': '2.0'}
        expected_row_2 = {'0': '01', '1': 'b', '2': '0.8333333333333334', '3': '1234567', '4': 'perfect',
                          '5': 'I love this product', '6': '5.0'}
        expected_row_3 = {'0': '01', '1': 'c', '2': '0.8666666666666667', '3': '1234567', '4': 'awesome',
                          '5': 'The perfect gift for my darling', '6': '4.0'}
        expected_row_4 = {'0': '02', '1': 'a', '2': '-0.3666666666666667', '3': '1234567', '4': 'a disaster',
                          '5': 'Too much expensive ', '6': '1.0'}
        expected_row_5 = {'0': '02', '1': 'c', '2': '0.6', '3': '1234567', '4': 'really good',
                          '5': 'A good compromise', '6': '3.5'}
        expected_row_6 = {'0': '03', '1': 'b', '2': '0.6666666666666666', '3': '1234567', '4': 'Awesome',
                          '5': '', '6': '5.0'}

        csv_iterator = iter(csv)

        result_row_1 = next(csv_iterator)
        result_row_2 = next(csv_iterator)
        result_row_3 = next(csv_iterator)
        result_row_4 = next(csv_iterator)
        result_row_5 = next(csv_iterator)
        result_row_6 = next(csv_iterator)

        with self.assertRaises(StopIteration):
            next(csv_iterator)

        self.assertDictEqual(expected_row_1, result_row_1)
        self.assertDictEqual(expected_row_2, result_row_2)
        self.assertDictEqual(expected_row_3, result_row_3)
        self.assertDictEqual(expected_row_4, result_row_4)
        self.assertDictEqual(expected_row_5, result_row_5)
        self.assertDictEqual(expected_row_6, result_row_6)

    def test__len_no_header(self):
        csv = CSVFile(csv_no_header, has_header=False)

        self.assertEqual(6, len(csv))

    def test_iter_tsv(self):

        tsv = CSVFile(tsv_file, has_header=False, separator='\t')

        expected_row_1 = {'0': 'listen', '1': 'improve', '2': 'differ'}
        expected_row_2 = {'0': 'visitor', '1': 'meant', '2': 'kind'}
        expected_row_3 = {'0': 'basis', '1': 'climb', '2': 'honor'}
        expected_row_4 = {'0': 'simple', '1': 'vote', '2': 'closer'}
        expected_row_5 = {'0': 'blind', '1': 'finger', '2': 'pencil'}
        expected_row_6 = {'0': 'clock', '1': 'energy', '2': 'shape'}

        tsv_iterator = iter(tsv)

        result_row_1 = next(tsv_iterator)
        result_row_2 = next(tsv_iterator)
        result_row_3 = next(tsv_iterator)
        result_row_4 = next(tsv_iterator)
        result_row_5 = next(tsv_iterator)
        result_row_6 = next(tsv_iterator)

        with self.assertRaises(StopIteration):
            next(tsv_iterator)

        self.assertDictEqual(expected_row_1, result_row_1)
        self.assertDictEqual(expected_row_2, result_row_2)
        self.assertDictEqual(expected_row_3, result_row_3)
        self.assertDictEqual(expected_row_4, result_row_4)
        self.assertDictEqual(expected_row_5, result_row_5)
        self.assertDictEqual(expected_row_6, result_row_6)

        self.assertEqual(6, len(tsv))


class TestJSONFile(TestCase):

    def test_iter(self):

        js = JSONFile(json_file)
        my_iter = iter(js)
        d1 = {"Title": "Jumanji", "Year": "1995", "Rated": "PG", "Released": "15 Dec 1995", "Runtime": "104 min",
              "Genre": "Adventure, Family, Fantasy", "Director": "Joe Johnston",
              "Writer": "Jonathan Hensleigh (screenplay by), Greg Taylor (screenplay by), Jim Strain (screenplay by), Greg Taylor (screen story by), Jim Strain (screen story by), Chris Van Allsburg (screen story by), Chris Van Allsburg (based on the book by)",
              "Actors": "Robin Williams, Jonathan Hyde, Kirsten Dunst, Bradley Pierce",
              "Plot": "After being trapped in a jungle board game for 26 years, a Man-Child wins his release from the game. But, no sooner has he arrived that he is forced to play again, and this time sets the creatures of the jungle loose on the city. Now it is up to him to stop them.",
              "Language": "English, French", "Country": "USA", "Awards": "4 wins & 9 nominations.",
              "Poster": "https://m.media-amazon.com/images/M/MV5BZTk2ZmUwYmEtNTcwZS00YmMyLWFkYjMtNTRmZDA3YWExMjc2XkEyXkFqcGdeQXVyMTQxNzMzNDI@._V1_SX300.jpg",
              "Ratings": [{"Source": "Internet Movie Database", "Value": "6.9/10"},
                          {"Source": "Rotten Tomatoes", "Value": "53%"}, {"Source": "Metacritic", "Value": "39/100"}],
              "Metascore": "39", "imdbRating": "6.9", "imdbVotes": "260,909", "imdbID": "tt0113497", "Type": "movie",
              "DVD": "25 Jan 2000", "BoxOffice": "N/A", "Production": "Sony Pictures Home Entertainment",
              "Website": "N/A",
              "Response": "True"}
        d2 = {"Title": "Grumpier Old Men", "Year": "1995", "Rated": "PG-13", "Released": "22 Dec 1995",
              "Runtime": "101 min",
              "Genre": "Comedy, Romance", "Director": "Howard Deutch",
              "Writer": "Mark Steven Johnson (characters), Mark Steven Johnson",
              "Actors": "Walter Matthau, Jack Lemmon, Sophia Loren, Ann-Margret",
              "Plot": "Things don't seem to change much in Wabasha County: Max and John are still fighting after 35 years, Grandpa still drinks, smokes, and chases women , and nobody's been able to catch the fabled \"Catfish Hunter\", a gigantic catfish that actually smiles at fishermen who try to snare it. Six months ago John married the new girl in town (Ariel), and people begin to suspect that Max might be missing something similar in his life. The only joy Max claims is left in his life is fishing, but that might change with the new owner of the bait shop.",
              "Language": "English, Italian, German", "Country": "USA", "Awards": "2 wins & 2 nominations.",
              "Poster": "https://m.media-amazon.com/images/M/MV5BMjQxM2YyNjMtZjUxYy00OGYyLTg0MmQtNGE2YzNjYmUyZTY1XkEyXkFqcGdeQXVyMTQxNzMzNDI@._V1_SX300.jpg",
              "Ratings": [{"Source": "Internet Movie Database", "Value": "6.6/10"},
                          {"Source": "Rotten Tomatoes", "Value": "17%"}, {"Source": "Metacritic", "Value": "46/100"}],
              "Metascore": "46", "imdbRating": "6.6", "imdbVotes": "21,823", "imdbID": "tt0113228", "Type": "movie",
              "DVD": "18 Nov 1997", "BoxOffice": "N/A", "Production": "Warner Home Video", "Website": "N/A",
              "Response": "True"}
        d3 = {"Title": "Toy Story", "Year": "1995", "Rated": "G", "Released": "22 Nov 1995", "Runtime": "81 min",
              "Genre": "Animation, Adventure, Comedy, Family, Fantasy", "Director": "John Lasseter",
              "Writer": "John Lasseter (original story by), Pete Docter (original story by), Andrew Stanton (original story by), Joe Ranft (original story by), Joss Whedon (screenplay by), Andrew Stanton (screenplay by), Joel Cohen (screenplay by), Alec Sokolow (screenplay by)",
              "Actors": "Tom Hanks, Tim Allen, Don Rickles, Jim Varney",
              "Plot": "A little boy named Andy loves to be in his room, playing with his toys, especially his doll named \"Woody\". But, what do the toys do when Andy is not with them, they come to life. Woody believes that he has life (as a toy) good. However, he must worry about Andy's family moving, and what Woody does not know is about Andy's birthday party. Woody does not realize that Andy's mother gave him an action figure known as Buzz Lightyear, who does not believe that he is a toy, and quickly becomes Andy's new favorite toy. Woody, who is now consumed with jealousy, tries to get rid of Buzz. Then, both Woody and Buzz are now lost. They must find a way to get back to Andy before he moves without them, but they will have to pass through a ruthless toy killer, Sid Phillips.",
              "Language": "English", "Country": "USA",
              "Awards": "Nominated for 3 Oscars. Another 23 wins & 17 nominations.",
              "Poster": "https://m.media-amazon.com/images/M/MV5BMDU2ZWJlMjktMTRhMy00ZTA5LWEzNDgtYmNmZTEwZTViZWJkXkEyXkFqcGdeQXVyNDQ2OTk4MzI@._V1_SX300.jpg",
              "Ratings": [{"Source": "Internet Movie Database", "Value": "8.3/10"},
                          {"Source": "Rotten Tomatoes", "Value": "100%"}, {"Source": "Metacritic", "Value": "95/100"}],
              "Metascore": "95", "imdbRating": "8.3", "imdbVotes": "761,649", "imdbID": "tt0114709", "Type": "movie",
              "DVD": "20 Mar 2001", "BoxOffice": "N/A", "Production": "Buena Vista",
              "Website": "http://www.disney.com/ToyStory", "Response": "True"}

        self.assertDictEqual(next(my_iter), d1)
        self.assertDictEqual(next(my_iter), d2)
        self.assertDictEqual(next(my_iter), d3)

    def test__len(self):

        dat = JSONFile(json_file)
        self.assertEqual(20, len(dat))


class TestDATFile(TestCase):
    def test_iter(self):

        dat = DATFile(dat_file)
        my_iter = iter(dat)

        expected = [
            {'0': '1', '1': 'F', '2': '1', '3': '10', '4': '48067'},
            {'0': '2', '1': 'M', '2': '56', '3': '16', '4': '70072'},
            {'0': '3', '1': 'M', '2': '25', '3': '15', '4': '55117'}
        ]
        for line in expected:
            dat1 = next(my_iter)
            self.assertEqual(line, dat1)

    def test__len(self):

        dat = DATFile(dat_file)
        self.assertEqual(70, len(dat))



from collections import defaultdict
from typing import List, Optional
import re

from clayrs.content_analyzer.ratings_manager.ratings import Interaction
from clayrs.recsys.content_based_algorithm.content_based_algorithm import ContentBasedAlgorithm
from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsIndex
from clayrs.recsys.content_based_algorithm.exceptions import NotPredictionAlg, OnlyNegativeItems, EmptyUserRatings


class IndexQuery(ContentBasedAlgorithm):
    """
    Class for the search engine recommender using an index.
    It firsts builds a query using the representation(s) specified of the positive items, then uses the mentioned query
    to do an actual search inside the index: every items will have a score of "closeness" in relation to the
    query, we use this score to rank every item.

    Just be sure to use textual representation(s) to build a significant query and to make a significant search!

    Examples:

        * Interested in only a field representation, classic tfidf similarity,
        $threshold = 3$ (Every item with rating $>= 3$ will be considered as positive)

        >>> from clayrs import recsys as rs
        >>> alg = rs.IndexQuery({"Plot": 0}, threshold=3)

        * Interested in multiple field representations of the items, BM25 similarity,
        $threshold = None$ (Every item with rating $>=$ mean rating of the user will be considered as positive)

        >>> alg = rs.IndexQuery(
        >>>                     item_field={"Plot": [0, "original_text"],
        >>>                                 "Genre": [0, 1],
        >>>                                 "Director": "preprocessed_text"},
        >>>                     classic_similarity=False,
        >>>                     threshold=3)


        !!! info

            After instantiating the IndexQuery algorithm, pass it in the initialization of
            a CBRS and the use its method to calculate ranking for single user or multiple users:

            Examples:

                >>> cbrs = rs.ContentBasedRS(algorithm=alg, ...)
                >>> cbrs.fit_rank(...)
                >>> # ...

    Args:
        item_field: dict where the key is the name of the field
            that contains the content to use, value is the representation(s) id(s) that will be
            used for the said item, just BE SURE to use textual representation(s). The value of a field can be a string
            or a list, use a list if you want to use multiple representations for a particular field.
        classic_similarity: True if you want to use the classic implementation of tfidf in Whoosh,
            False if you want BM25F
        threshold: Threshold for the ratings. If the rating is greater than the threshold, it will be considered
            as positive. If the threshold is not specified, the average score of all items liked by the user is used.
    """
    __slots__ = ('_string_query', '_scores', '_positive_user_docs', '_classic_similarity')

    def __init__(self, item_field: dict, classic_similarity: bool = True, threshold: float = None):
        super().__init__(item_field, threshold)
        self._string_query: Optional[str] = None
        self._scores: Optional[list] = None
        self._positive_user_docs: Optional[dict] = None
        self._classic_similarity: bool = classic_similarity

    def _get_representations(self, index_representations: dict):
        """
        Private method which extracts representation(s) chosen from all representations codified for the items
        extracted from the index

        Args:
            index_representations (dict): representations for an item extracted from the index
        """
        def find_valid(pattern: str):
            field_index_retrieved = [field_index for field_index in index_representations
                                     if re.match(pattern, field_index)]

            if len(field_index_retrieved) == 0:
                raise KeyError("Id {} not found for the field {}".format(id, k))
            elif len(field_index_retrieved) > 1:
                raise ValueError("This shouldn't happen! Duplicate fields?")
            else:
                valid = field_index_retrieved[0]

            return valid

        representations_valid = {}
        for k in self.item_field:
            for id in self.item_field[k]:
                # every representation for an item is codified like this: plot#0#tfidf
                if isinstance(id, str):
                    pattern = "^{}#.+#{}$".format(k, id)
                else:
                    # the id passed it's a int
                    pattern = "^{}#{}.*$".format(k, id)

                valid_key = find_valid(pattern)
                representations_valid[valid_key] = index_representations[valid_key]

        return representations_valid

    def _load_available_contents(self, index_path: str, items_to_load: set = None):
        return LoadedContentsIndex(index_path)

    def process_rated(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsIndex):
        """
        Function that extracts features from positive rated items ONLY!
        The extracted features will be used to fit the algorithm (build the query).

        Features extracted will be stored in private attributes of the class.

        Args:
            user_ratings: List of Interaction objects for a single user
            available_loaded_items: The LoadedContents interface which contains loaded contents
        """
        # a list since there could be duplicate interaction (eg bootstrap partitioning)
        items_scores_dict = defaultdict(list)
        for interaction in user_ratings:
            items_scores_dict[interaction.item_id].append(interaction.score)

        items_scores_dict = dict(sorted(items_scores_dict.items()))  # sort dictionary based on key for reproducibility

        threshold = self.threshold
        if threshold is None:
            threshold = self._calc_mean_user_threshold(user_ratings)

        # Initializes positive_user_docs which is a list that has tuples with document_id as first element and
        # a dictionary as second. The dictionary value has the name of the field as key
        # and its contents as value. By doing so we obtain the data of the fields while
        # also storing information regarding the field and the document where it was
        scores = []
        positive_user_docs = []

        ix = available_loaded_items.get_contents_interface()

        # we extract feature of each item sorted based on its key: IMPORTANT for reproducibility!!
        for item_id, score_list in items_scores_dict.items():

            score_assigned = map(float, score_list)

            for score in score_assigned:
                if score >= threshold:
                    # {item_id: {"item": item_dictionary, "score": item_score}}
                    item_query = ix.query(item_id, results_number=1, classic_similarity=self._classic_similarity)
                    if len(item_query) != 0:
                        item = item_query.pop(item_id).get('item')
                        scores.append(score)
                        positive_user_docs.append((item_id, self._get_representations(item)))

        if len(user_ratings) == 0:
            raise EmptyUserRatings("The user selected doesn't have any ratings!")

        user_id = user_ratings[0].user_id
        if len(positive_user_docs) == 0:
            raise OnlyNegativeItems(f"User {user_id} - There are no rated items available locally or there are only "
                                    f"negative items available locally!")

        self._positive_user_docs = positive_user_docs
        self._scores = scores

    def fit(self):
        """
        The fit process for the IndexQuery consists in building a query using the features of the positive items ONLY
        (items that the user liked). The terms relative to these 'positive' items are boosted by the
        rating he/she gave.

        This method uses extracted features of the positive items stored in a private attribute, so
        process_rated() must be called before this method.

        The built query will also be stored in a private attribute.
        """
        # For each field of each document one string (containing the name of the field and the data in it)
        # is created and added to the query.
        # Also each part of the query that refers to a document
        # is boosted by the score given by the user to said document
        string_query = "("
        for (doc_id, doc_data), score in zip(self._positive_user_docs, self._scores):
            string_query += "("
            for field_name in doc_data:
                if field_name == 'content_id':
                    continue
                word_list = doc_data[field_name].split()
                string_query += field_name + ":("
                for term in word_list:
                    string_query += term + " "
                string_query += ") "
            string_query += ")^" + str(score) + " "
        string_query += ") "

        self._string_query = string_query

    def _build_mask_list(self, user_seen_items: set, filter_list: List[str] = None):
        """
        Private function that calculate the mask query and the filter query for whoosh to use:

        - The mask query is needed to ignore items already rated by the user
        - The filter query is needed to predict only items present in the filter_list

        If in the filter list there are items already rated by the user, those are excluded in the
        mask query so that the prediction for those items can be calculated

        Args:
            user_seen_items: set of items present in the user profile
            filter_list: list of the items to predict, if None all unrated items will be predicted
        """
        masked_list = list(user_seen_items)
        if filter_list is not None:
            masked_list = [item for item in user_seen_items if item not in filter_list]

        return masked_list

    def predict(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsIndex,
                filter_list: List[str] = None) -> List[Interaction]:
        """
        IndexQuery is not a Prediction Score Algorithm, so if this method is called,
        a NotPredictionAlg exception is raised

        Raises:
            NotPredictionAlg: exception raised since the CentroidVector algorithm is not a score prediction algorithm
        """
        raise NotPredictionAlg("IndexQuery is not a Score Prediction Algorithm!")

    def rank(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsIndex,
             recs_number: int = None, filter_list: List[str] = None) -> List[Interaction]:
        """
        Rank the top-n recommended items for the user. If the recs_number parameter isn't specified,
        All unrated items for the user will be ranked (or only items in the filter list, if specified).

        One can specify which items must be ranked with the `filter_list` parameter,
        in this case ONLY items in the `filter_list` parameter will be ranked.
        One can also pass items already seen by the user with the filter_list parameter.
        Otherwise, **ALL** unrated items will be ranked.

        Args:
            user_ratings: List of Interaction objects for a single user
            available_loaded_items: The LoadedContents interface which contains loaded contents. In this case it will
                be a `LoadedContentsIndex`
            recs_number: number of the top ranked items to return, if None all ranked items will be returned
            filter_list (list): list of the items to rank, if None all unrated items for the user will be ranked

        Returns:
            List of Interactions object in a descending order w.r.t the 'score' attribute, representing the ranking for
                a single user
        """
        try:
            user_id = user_ratings[0].user_id
        except IndexError:
            raise EmptyUserRatings("The user selected doesn't have any ratings!")

        user_seen_items = set([interaction.item_id for interaction in user_ratings])

        mask_list = self._build_mask_list(user_seen_items, filter_list)

        ix = available_loaded_items.get_contents_interface()
        score_docs = ix.query(self._string_query, recs_number, mask_list, filter_list, self._classic_similarity)

        # we construct the output data
        rank_interaction_list = [Interaction(user_id, item_id, score_docs[item_id]['score'])
                                 for item_id in score_docs]

        return rank_interaction_list

    def __str__(self):
        return "IndexQuery"

    def __repr__(self):
        return f'IndexQuery(item_field={self.item_field}, classic_similarity={self._classic_similarity}, ' \
               f'threshold={self.threshold})'



from . import centroid_vector
from . import classifier
from . import index_query
from . import regressor

from .centroid_vector import *
from .classifier import *
from .index_query import *
from .regressor import *



from abc import ABC, abstractmethod
from collections import defaultdict
from typing import List

import pandas as pd
from scipy.stats import ttest_ind, ranksums


class StatisticalTest(ABC):
    """
    Abstract class for Statistical Test.

    Every statistical test have to identify common users if the user
    chooses to pass us a df. The method stat_test_results is implemented
    differently for each statistical test you decide to do.
    """
    @staticmethod
    def _common_users(df1, df2, column_list) -> pd.DataFrame:
        """
        Method called by the statistical test in case of use of df.
        Common users are searched for meaningful comparison.
        """
        # we need to also extract the user_id column, that's why we append 'user_id'
        column_list.append('user_id')

        df1 = df1[column_list]
        df2 = df2[column_list]

        common_rows = pd.merge(df1, df2, how="inner", on=['user_id'])

        return common_rows

    @abstractmethod
    def perform(self, users_metric_results: list) -> pd.DataFrame:
        """
        Abstract method in which must be specified how to calculate Statistical test
        """
        pass


class PairedTest(StatisticalTest):

    def perform(self, df_list: List[pd.DataFrame]) -> pd.DataFrame:
        """
        Method which performs the chosen paired statistical test.

        Since it's a paired test, the final result is a pandas DataFrame which contains learning
        schemas compared in pair.
        For example if you call the `perform()` method by passing a list containing three different DataFrames, one for
        each learning schema to compare:

        ```python
        # Ttest as example since it's a Paired Test
        Ttest().perform([user_df1, user_df2, user_df3])
        ```

        You will obtain a DataFrame comparing all different combinations:

        * (system1, system2)
        * (system1, system3)
        * (system2, system3)

        The first value of each cell is the ***statistic***, the second is the ***p-value***

        Args:
            df_list: List containing DataFrames with several metrics to compare, preferably metrics computed for each
                user. One DataFrame corresponds to one learning schema

        Returns:
            A Pandas DataFrame where each combination of learning schemas are compared in pair. The first value of each
                cell is the ***statistic***, the second is the ***p-value***

        """

        # we consider the 'user_id' index as a column
        df_list = [df.reset_index() if 'user_id' not in df.columns else df for df in df_list]

        final_result = defaultdict(list)

        n_system_evaluated = 1
        while len(df_list) != 0:
            df1 = df_list.pop(0)
            for i, other_df in enumerate(df_list, start=n_system_evaluated + 1):
                common_metrics = [column for column in df1.columns
                                  if column != 'user_id' and column in other_df.columns]

                common_rows = self._common_users(df1, other_df, list(common_metrics))

                final_result["Systems evaluated"].append((f"system_{n_system_evaluated}", f"system_{i}"))
                for metric in common_metrics:

                    # drop nan values since otherwise test may behave unexpectedly
                    metric_rows = common_rows[[f"{metric}_x", f"{metric}_y"]].dropna()

                    score_system1 = metric_rows[f"{metric}_x"]
                    score_system2 = metric_rows[f"{metric}_y"]

                    single_metric_result = self._perform_test(score_system1, score_system2)
                    final_result[metric].append(single_metric_result)

            n_system_evaluated += 1

        return pd.DataFrame(final_result).set_index("Systems evaluated")

    @abstractmethod
    def _perform_test(self, score_metric_system1: list, score_metric_system2: list):
        """
        Abstract method in which must be specified how to calculate the paired statistical test
        """
        raise NotImplementedError


class Ttest(PairedTest):
    """
    Calculate the T-test for the means of *two independent* samples of scores.

    This is a two-sided test for the null hypothesis that 2 independent samples
    have identical average (expected) values. This test assumes that the
    populations have identical variances by default.
    """
    def _perform_test(self, score_metric_system1: list, score_metric_system2: list):
        return ttest_ind(score_metric_system1, score_metric_system2)


class Wilcoxon(PairedTest):
    """
    Compute the Wilcoxon rank-sum statistic for two samples.

    The Wilcoxon rank-sum test tests the null hypothesis that two sets
    of measurements are drawn from the same distribution. The alternative
    hypothesis is that values in one sample are more likely to be
    larger than the values in the other sample.
    """
    def _perform_test(self, score_metric_system1: list, score_metric_system2: list):
        return ranksums(score_metric_system1, score_metric_system2)



from unittest import TestCase

from clayrs.content_analyzer.exogenous_properties_retrieval import PropertiesFromDataset

from clayrs.content_analyzer import FieldConfig, ExogenousConfig


class TestFieldConfig(TestCase):
    def test_invalid_id(self):
        with self.assertRaises(ValueError):
            FieldConfig(id='.in.vali.d')

        with self.assertRaises(ValueError):
            FieldConfig(id='#in#vali#d')

        with self.assertRaises(ValueError):
            FieldConfig(id='     ')

        with self.assertRaises(ValueError):
            FieldConfig(id='is invalid')

        with self.assertRaises(ValueError):
            FieldConfig(id='is/inva/lid')

        # ...and many more

    def test_valid_id(self):
        valid_object = FieldConfig(id='test')
        self.assertIsNotNone(valid_object)

        valid_object = FieldConfig(id='test_valid')
        self.assertIsNotNone(valid_object)

        valid_object = FieldConfig(id='test-valid')
        self.assertIsNotNone(valid_object)

        valid_object = FieldConfig(id='test1-valid2')
        self.assertIsNotNone(valid_object)

        valid_object = FieldConfig(id='1_2-3_')
        self.assertIsNotNone(valid_object)

        # ...and many more


class TestExogenousConfig(TestCase):
    def test_invalid_id(self):
        with self.assertRaises(ValueError):
            ExogenousConfig(exogenous_technique=PropertiesFromDataset(), id='.in.vali.d')

        with self.assertRaises(ValueError):
            ExogenousConfig(exogenous_technique=PropertiesFromDataset(), id='.in.vali.d')

        with self.assertRaises(ValueError):
            ExogenousConfig(exogenous_technique=PropertiesFromDataset(), id='.in.vali.d')

        with self.assertRaises(ValueError):
            ExogenousConfig(exogenous_technique=PropertiesFromDataset(), id='.in.vali.d')

        with self.assertRaises(ValueError):
            ExogenousConfig(exogenous_technique=PropertiesFromDataset(), id='.in.vali.d')

        # ...and many more

    def test_valid_id(self):
        valid_object = ExogenousConfig(exogenous_technique=PropertiesFromDataset(), id='test')
        self.assertIsNotNone(valid_object)

        valid_object = ExogenousConfig(exogenous_technique=PropertiesFromDataset(), id='test_valid')
        self.assertIsNotNone(valid_object)

        valid_object = ExogenousConfig(exogenous_technique=PropertiesFromDataset(), id='test-valid')
        self.assertIsNotNone(valid_object)

        valid_object = ExogenousConfig(exogenous_technique=PropertiesFromDataset(), id='test1-valid2')
        self.assertIsNotNone(valid_object)

        valid_object = ExogenousConfig(exogenous_technique=PropertiesFromDataset(), id='1_2-3_')
        self.assertIsNotNone(valid_object)

        # ...and many more



from __future__ import annotations
from typing import List, Union, Tuple, TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from clayrs.content_analyzer.ratings_manager.ratings import Prediction, Rank, Ratings
    from clayrs.evaluation.metrics.metrics import Metric

from clayrs.evaluation.eval_pipeline_modules.metric_evaluator import MetricEvaluator
from clayrs.utils.const import logger


class EvalModel:
    """
    Class for evaluating a recommender system.

    The Evaluation module needs the following parameters:

    *   A list of computed rank/predictions (in case multiple splits must be evaluated)
    *   A list of truths (in case multiple splits must be evaluated)
    *   List of metrics to compute

    Obviously the list of computed rank/predictions and list of truths must have the same length,
    and the rank/prediction in position $i$ will be compared with the truth at position $i$

    Examples:

        >>> import clayrs.evaluation as eva
        >>>
        >>> em = eva.EvalModel(
        >>>         pred_list=rank_list,
        >>>         truth_list=truth_list,
        >>>         metric_list=[
        >>>             eva.NDCG(),
        >>>             eva.Precision()
        >>>             eva.RecallAtK(k=5, sys_average='micro')
        >>>         ]
        >>> )

    Then call the fit() method of the instantiated EvalModel to perform the actual evaluation

    Args:
        pred_list: Recommendations list to evaluate. It's a list in case multiple splits must be evaluated. Both Rank
            objects (where items are ordered and the score is not relevant) or Prediction objects (where the score
             predicted is the predicted rating for the user regarding a certain item) can be evaluated
        truth_list: Ground truths list used to compare recommendations. It's a list in case multiple splits must
            be evaluated.
        metric_list: List of metrics that will be used to evaluate recommendation list specified

    Raises:
        ValueError: ValueError is raised in case the pred_list and truth_list are empty or have different length
    """
    def __init__(self,
                 pred_list: Union[List[Prediction], List[Rank]],
                 truth_list: List[Ratings],
                 metric_list: List[Metric]):

        if len(pred_list) == 0 and len(truth_list) == 0:
            raise ValueError("List containing predictions and list containing ground truths are empty!")
        elif len(pred_list) != len(truth_list):
            raise ValueError("List containing predictions and list containing ground truths must have the same length!")

        self._pred_list = pred_list
        self._truth_list = truth_list
        self._metric_list = metric_list

        self._yaml_report_result = None

    @property
    def pred_list(self) -> Union[List[Prediction], List[Rank]]:
        """
        List containing recommendations frame

        Returns:
            The list containing recommendations frame
        """
        return self._pred_list

    @property
    def truth_list(self) -> List[Ratings]:
        """
        List containing ground truths

        Returns:
            The list containing ground truths
        """
        return self._truth_list

    @property
    def metric_list(self) -> List[Metric]:
        """
        List of metrics used to evaluate recommendation lists

        Returns:
            The list containing all metrics
        """
        return self._metric_list

    def append_metric(self, metric: Metric):
        """
        Append a metric to the metric list that will be used to evaluate recommendation lists

        Args:
            metric: Metric to append to the metric list
        """
        self._metric_list.append(metric)

    def fit(self, user_id_list: list = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        This method performs the actual evaluation of the recommendation frames passed as input in the constructor of
        the class

        In case you want to perform evaluation for selected users, specify their ids parameter of this method.
        Otherwise, all users in the recommendation frames will be considered in the evaluation process

        Examples:

            >>> import clayrs.evaluation as eva
            >>> selected_users = ['u1', 'u22', 'u3'] # (1)
            >>> em = eva.EvalModel(
            >>>         pred_list,
            >>>         truth_list,
            >>>         metric_list=[eva.Precision(), eva.Recall()]
            >>> )
            >>> em.fit(selected_users)

        The method returns two pandas DataFrame: one containing ***system results*** for every metric in the metric
        list, one containing ***users results*** for every metric eligible

        Returns:
            The first DataFrame contains the **system result** for every metric inside the metric_list

            The second DataFrame contains every **user results** for every metric eligible inside the metric_list
        """
        logger.info('Performing evaluation on metrics chosen')

        pred_list = self._pred_list
        truth_list = self._truth_list

        if user_id_list is not None:
            user_id_list_set = set([str(user_id) for user_id in user_id_list])

            pred_list = [pred.filter_ratings(user_id_list_set) for pred in self._pred_list]
            truth_list = [truth.filter_ratings(user_id_list_set) for truth in self._truth_list]

        sys_result, users_result = MetricEvaluator(pred_list, truth_list).eval_metrics(self.metric_list)

        # we save the sys result for report yaml
        self._yaml_report_result = sys_result.to_dict(orient='index')

        return sys_result, users_result

    def __repr__(self):
        return f'EvalModel(pred_list={self._pred_list}, truth_list={self._truth_list},' \
               f' metric_list={self._metric_list}'



from . import content_representation
from . import embeddings
from . import field_content_production_techniques
from . import information_processor
from . import memory_interfaces
from . import ratings_manager

from .content_representation import *
from .embeddings import *
from .field_content_production_techniques import *
from .information_processor import *
from .memory_interfaces import *
from .ratings_manager import *
from .config import ExogenousConfig, UserAnalyzerConfig, ItemAnalyzerConfig, FieldConfig
from .content_analyzer_main import ContentAnalyzer
from .exogenous_properties_retrieval import DBPediaMappingTechnique, PropertiesFromDataset, BabelPyEntityLinking
from .raw_information_source import CSVFile, JSONFile, DATFile, SQLDatabase





import os
from unittest import TestCase
import pandas as pd

from clayrs.content_analyzer.ratings_manager.ratings import Rank, Ratings
from clayrs.evaluation import MAP
from clayrs.evaluation.metrics.classification_metrics import Precision, Recall

from clayrs.evaluation.eval_model import MetricEvaluator
from clayrs.evaluation.metrics.plot_metrics import LongTailDistr, PopRecsCorrelation


# Every Metric is tested singularly, so we just check that everything goes smoothly at the
# MetricEvaluator level
class TestMetricEvaluator(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        original_ratings = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1',
                         'u2', 'u2', 'u2', 'u2',
                         'u3', 'u3', 'u3', 'u3', 'u3', 'u3',
                         'u4', 'u4', 'u4', 'u4', 'u4', 'u4', 'u4',
                         'u5', 'u5', 'u5', 'u5', 'u5',
                         'u6', 'u6'],
             'item_id': ['i1', 'i2', 'i3', 'i4', 'i5', 'i6', 'i7', 'i8',
                         'i1', 'i9', 'i10', 'i11',
                         'i1', 'i12', 'i13', 'i3', 'i10', 'i14',
                         'i3', 'i10', 'i15', 'i16', 'i9', 'i17', 'i99',
                         'i10', 'i18', 'i19', 'i20', 'i21',
                         'inew_1', 'inew_2'],
             'score': [5, 4, 4, 1, 2, 3, 3, 1,
                       4, 5, 1, 1,
                       3, 3, 2, 1, 1, 4,
                       4, 4, 5, 5, 3, 3, 3,
                       3, 3, 2, 2, 1,
                       4, 3]})
        cls.original_ratings = Ratings.from_dataframe(original_ratings)

        train = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1',  # removed last 2
                         'u2', 'u2', 'u2',  # removed last 1
                         'u3', 'u3', 'u3', 'u3',  # removed last 2
                         'u4', 'u4', 'u4', 'u4', 'u4',  # removed last 2
                         'u5', 'u5', 'u5', 'u5',  # removed last 1
                         'u6'],  # removed last 1
             'item_id': ['i1', 'i2', 'i3', 'i4', 'i5', 'i6',
                         'i1', 'i9', 'i10',
                         'i1', 'i12', 'i13', 'i3',
                         'i3', 'i10', 'i15', 'i16', 'i9',
                         'i10', 'i18', 'i19', 'i20',
                         'inew_1'],
             'score': [5, 4, 4, 1, 2, 3,
                       4, 5, 1,
                       3, 3, 2, 1,
                       4, 4, 5, 5, 3,
                       3, 3, 2, 2,
                       4]})
        cls.train = Ratings.from_dataframe(train)

        truth = pd.DataFrame(
            {'user_id': ['u1', 'u1',
                         'u2',
                         'u3', 'u3',
                         'u4', 'u4', 'u4',
                         'u5',
                         'u6'],
             'item_id': ['i7', 'i8',
                         'i11',
                         'i10', 'i14',
                         'i9', 'i17', 'i99',
                         'i21',
                         'inew_2'],
             'score': [3, 1,
                       1,
                       1, 4,
                       3, 3, 3,
                       1,
                       3]})
        truth = Ratings.from_dataframe(truth)

        # u6 is missing, just to test DeltaGap in case for some users recs can't be computed
        recs = pd.DataFrame(
            {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1',
                         'u2', 'u2', 'u2', 'u2', 'u2',
                         'u3', 'u3', 'u3', 'u3', 'u3',
                         'u4', 'u4', 'u4', 'u4', 'u4',
                         'u5', 'u5', 'u5', 'u5', 'u5'],
             'item_id': ['i7', 'i10', 'i11', 'i12', 'i13',
                         'i11', 'i20', 'i6', 'i3', 'i4',
                         'i4', 'i5', 'i6', 'i7', 'i10',
                         'i9', 'i2', 'i3', 'i1', 'i5',
                         'i2', 'i3', 'i4', 'i5', 'i6'],
             'score': [500, 400, 300, 200, 100,
                       400, 300, 200, 100, 50,
                       150, 125, 110, 100, 80,
                       390, 380, 360, 320, 200,
                       250, 150, 190, 100, 50]})
        recs = Rank.from_dataframe(recs)

        cls.rank_pred_list = [recs]
        cls.truth_list = [truth]

    def test_eval_metrics_empty_dfs(self):
        # test eval_metrics with metrics which returns empty dataframe
        metric_list = [PopRecsCorrelation(self.original_ratings), LongTailDistr()]

        sys_result, users_results = MetricEvaluator(self.rank_pred_list, self.truth_list).eval_metrics(metric_list)

        self.assertTrue(len(sys_result) == 0)
        self.assertTrue(len(users_results) == 0)

    def test_eval_metrics_users_missing_truth(self):
        rank_wo_u3 = pd.DataFrame({
            'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1',
                        'u2', 'u2', 'u2', 'u2', 'u2', 'u2', 'u2'],
            'item_id': ['i9', 'i6', 'inew1', 'inew2', 'i2', 'i1', 'i8',
                        'i10', 'inew3', 'i2', 'i1', 'i8', 'i4', 'i9'],

            'score': [500, 450, 400, 350, 300, 200, 150,
                      400, 300, 200, 100, 50, 25, 10]
        })
        rank_wo_u3 = Rank.from_dataframe(rank_wo_u3)

        truth = pd.DataFrame({
            'user_id': ['u1', 'u1', 'u1', 'u1', 'u1',
                        'u2', 'u2', 'u2', 'u2', 'u2',
                        'u3', 'u3', 'u3', 'u3', 'u3'],
            'item_id': ['i1', 'i2', 'i6', 'i8', 'i9',
                        'i1', 'i2', 'i4', 'i9', 'i10',
                        'i2', 'i3', 'i12', 'imissing3', 'imissing4'],

            'score': [3, 3, 4, 1, 1,
                      5, 3, 3, 4, 4,
                      4, 2, 3, 3, 3]
        })
        truth = Ratings.from_dataframe(truth)

        rank_list = [rank_wo_u3]
        truth_list = [truth]

        sys_result, users_results = MetricEvaluator(rank_list, truth_list).eval_metrics([Precision(), Recall(), MAP()])

        # check that u3 isn't present in results since we don't have any prediction for it
        self.assertEqual({'u1', 'u2'}, set(users_results.index))

        # the user result frame must contain results for each user of the Precision, Recall and AP
        self.assertEqual(list(users_results.columns), ['Precision - macro', 'Recall - macro', 'AP'])

        # the sys_result frame must contain result of the system for each fold (1 in this case) + the mean result
        self.assertTrue(len(sys_result) == 2)
        self.assertEqual({'sys - fold1', 'sys - mean'}, set(sys_result.index))

        # the sys result frame must contain results for the system of the Precision, Recall and MAP
        self.assertEqual(list(sys_result.columns), ['Precision - macro', 'Recall - macro', 'MAP'])

    @classmethod
    def tearDownClass(cls) -> None:
        os.remove('long_tail_distr_truth.png')
        os.remove('pop_recs_correlation.png')
        os.remove('pop_recs_correlation_no_zeros.png')



from __future__ import annotations
import gc
import json
import pickle
import re
import lzma
import os
import shutil

from typing import List, Dict, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.content_analyzer.config import ContentAnalyzerConfig
    from clayrs.content_analyzer.memory_interfaces.memory_interfaces import InformationInterface

from clayrs.content_analyzer.content_representation.content import Content, IndexField, ContentEncoder
from clayrs.utils.const import logger
from clayrs.utils.context_managers import get_progbar
from clayrs.content_analyzer.utils.id_merger import id_merger


class ContentAnalyzer:
    """
    Class to whom the control of the content analysis phase is delegated. It uses the data stored in the configuration
    file to create and serialize the contents the user wants to produce. It also checks that the configurations the
    user wants to run on the raw contents have unique ids (otherwise it would be impossible to refer to a particular
    field representation or exogenous representation)
    Args:
        config (ContentAnalyzerConfig): configuration for processing the item fields. This parameter provides
            the possibility of customizing the way in which the input data is processed.
    """

    def __init__(self, config: ContentAnalyzerConfig):
        self._config: ContentAnalyzerConfig = config

    def set_config(self, config: ContentAnalyzerConfig):
        self._config = config

    def fit(self):
        """
        Processes the creation of the contents and serializes the contents. This method starts the content production
        process and initializes everything that will be used to create said contents, their fields and their
        representations
        """
        # before starting the process, the content analyzer manin checks that there are no duplicate id cases
        # both in the field dictionary and in the exogenous representation list
        # this is done now and not recursively for each content during the creation process, in order to avoid starting
        # an operation that is going to fail
        try:
            self.__check_field_dict()
            self.__check_exogenous_representation_list()
        except ValueError as e:
            raise e

        # creates the directory where the data will be serialized and overwrites it if it already exists
        output_path = self._config.output_directory
        if os.path.exists(output_path):
            shutil.rmtree(output_path)
        os.makedirs(output_path)

        contents_producer = ContentsProducer.get_instance()
        contents_producer.set_config(self._config)
        created_contents = contents_producer.create_contents()

        if self._config.export_json:
            json_path = os.path.join(self._config.output_directory, 'contents.json')
            with open(json_path, "w") as data:
                json.dump(created_contents, data, cls=ContentEncoder, indent=4)

        with get_progbar(created_contents) as pbar:
            pbar.set_description("Serializing contents")

            for content in pbar:
                self.__serialize_content(content)

    def __serialize_content(self, content: Content):
        """
        This method serializes a specific content in the output directory defined by the content analyzer config
        Args:
            content (Content): content instance that will be serialized
        """

        file_name = re.sub(r'[^\w\s]', '', content.content_id)
        path = os.path.join(self._config.output_directory, file_name + '.xz')
        with lzma.open(path, 'wb') as f:
            pickle.dump(content, f, protocol=4)

    def __check_field_dict(self):
        """
        This function checks that there are no duplicate ids in the field_dict for a specific field_name.
        If this is not the case and a duplicate is found, a ValueError exception is thrown to warn the user.
        If the config id is None, the representation name will be kept as None even if it is not unique.
        So any case where the id is None is not considered.
        """
        for field_name in self._config.get_field_name_list():
            custom_id_list = [config.id for config in self._config.get_configs_list(field_name)
                              if config.id is not None]

            if len(custom_id_list) != len(set(custom_id_list)):
                raise ValueError("Each id for each field config of the field %s has to be unique" % field_name)

    def __check_exogenous_representation_list(self):
        """
        This function checks that there are no duplicate ids in the exogenous_representation_list.
        If this is not the case and a duplicate is found, a ValueError exception is thrown to warn the user.
        If the config id is None, the representation name will be kept as None even if it is not unique.
        So any case where the id is None is not considered
        """
        custom_id_list = [config.id for config in self._config.exogenous_representation_list if config.id is not None]
        if len(custom_id_list) != len(set(custom_id_list)):
            raise ValueError("Each id for each exogenous config in the exogenous list has to be unique")

    def __str__(self):
        return "ContentAnalyzer"

    def __repr__(self):
        return f'ContentAnalyzer(config={self._config})'


class ContentsProducer:
    """
    Singleton class which encapsulates the creation process of the items,
    The creation process is specified in the config parameter of ContentAnalyzer and
    is supposed to be the same for each item.
    """
    __instance = None

    @staticmethod
    def get_instance():
        """
        returns the singleton instance
        Returns:
            ContentsProducer: instance
        """
        # Static access method
        if ContentsProducer.__instance is None:
            ContentsProducer.__instance = ContentsProducer()
        return ContentsProducer.__instance

    def __init__(self):
        self.__config: ContentAnalyzerConfig = None
        # dictionary of memory interfaces defined in the FieldConfigs. The key is the directory of the memory interface
        # and the value is the memory interface itself (only one memory interface can be defined for each directory)
        # if a memory interface has an already defined directory, the memory interface associated to said directory
        # will be considered instead
        self.__memory_interfaces: Dict[InformationInterface] = {}
        # Virtually private constructor.
        if ContentsProducer.__instance is not None:
            raise Exception("This class is a singleton!")
        ContentsProducer.__instance = self

    def set_config(self, config: ContentAnalyzerConfig):
        self.__config = config

    def create_contents(self) -> List[Content]:
        """
        Creates the contents based on the information defined in the Content Analyzer's config
        Returns:
            contents_list (List[Content]): list of contents created by the method
        """
        if self.__config is None:
            raise Exception("You must set a config with set_config()")

        # will store the contents and is the variable that will be returned by the method
        contents_list = []

        for raw_content in self.__config.source:
            # construct id from the list of the fields that compound id
            content_id = id_merger(raw_content, self.__config.id)
            contents_list.append(Content(content_id))

        # two lists are instantiated, one for the configuration names (given by the user) and one for the exogenous
        # properties representations. These lists will maintain the data for the content creation. This is done
        # because otherwise it would be necessary to append directly to the content. But in the Content class
        # the representations are kept as dataframes and appending to dataframes is computationally heavy
        for ex_config in self.__config.exogenous_representation_list:
            lod_properties = ex_config.exogenous_technique.get_properties(self.__config.source)

            for i in range(len(contents_list)):
                contents_list[i].append_exogenous_representation(lod_properties[i], ex_config.id)

        # this dictionary will store any representation list that will be kept in one of the the index
        # the elements will be in the form:
        #   { memory_interface: {'Plot_0': [FieldRepr for content1, FieldRepr for content2, ...]}}
        # the 0 after the Plot field name is used to define the representation number associated with the Plot field
        # since it's possible to store multiple Plot fields in the index
        index_representations_dict = {}

        for field_name in self.__config.get_field_name_list():
            logger.info(f"   Processing field: {field_name}   ".center(50, '*'))

            for repr_number, field_config in enumerate(self.__config.get_configs_list(field_name)):

                # technique_result is a list of field representation produced by the content technique
                # each field repr in the list will refer to a content
                # technique_result[0] -> contents_list[0]
                technique_result = field_config.content_technique.produce_content(
                    field_name, field_config.preprocessing, self.__config.source)

                if field_config.memory_interface is not None:
                    memory_interface = field_config.memory_interface
                    # if the index for the directory in the config hasn't been defined yet in the contents producer,
                    # the index associated to the field config that is being processed is added to the
                    # contents producer's memory interfaces list, and will be used for the future field configs with
                    # an assigned memory interface that has the same directory.
                    # This means that only the index defined in the first FieldConfig that has one will actually be used
                    if memory_interface not in self.__memory_interfaces.values():
                        self.__memory_interfaces[memory_interface.directory] = memory_interface
                        index_representations_dict[memory_interface] = {}
                    else:
                        memory_interface = self.__memory_interfaces[memory_interface.directory]

                    if field_config.id is not None:
                        index_field_name = "{}#{}#{}".format(field_name, str(repr_number), field_config.id)
                    else:
                        index_field_name = "{}#{}".format(field_name, str(repr_number))

                    index_representations_dict[memory_interface][index_field_name] = technique_result

                    # in order to refer to the representation that will be stored in the index, an IndexField repr will
                    # be added to each content (and it will contain all the necessary information to retrieve the data
                    # from the index)
                    technique_result = [IndexField(index_field_name, i, memory_interface)
                                        for i in range(len(self.__config.source))]

                for i in range(len(contents_list)):
                    contents_list[i].append_field_representation(field_name, technique_result[i], field_config.id)

                del technique_result
                gc.collect()

        # after the contents creation process, the data to be indexed will be serialized inside of the memory interfaces
        # for each created content, a new entry in each index will be created
        # the entry will be in the following form: {"content_id": id, "Plot_0": "...", "Plot_1": "...", ...}
        if len(self.__memory_interfaces) != 0:
            for memory_interface in self.__memory_interfaces.values():
                memory_interface.init_writing(True)
                for i in range(0, len(contents_list)):
                    memory_interface.new_content()
                    memory_interface.new_field("content_id", contents_list[i].content_id)
                    for field_name in index_representations_dict[memory_interface].keys():
                        memory_interface.new_field(
                            field_name, str(index_representations_dict[memory_interface][field_name][i].value))
                    memory_interface.serialize_content()
                memory_interface.stop_writing()
            self.__memory_interfaces.clear()

        return contents_list

    def __str__(self):
        return "ContentsProducer"

    def __repr__(self):
        return f'ContentsProducer()'



from typing import List
import spacy
from spacy.tokens import Token

from clayrs.content_analyzer.information_processor.information_processor import NLP
from clayrs.content_analyzer.utils.check_tokenization import check_not_tokenized


class Spacy(NLP):
    """
    Interface to the Spacy library for natural language processing features

    Examples:

        * Strip multiple whitespaces from running text
        >>> spacy_obj = Spacy(strip_multiple_whitespaces=True)
        >>> spacy_obj.process('This   has  a lot  of   spaces')
        ['This', 'has', 'a', 'lot', 'of', 'spaces']

        * Remove punctuation from running text
        >>> spacy_obj = Spacy(remove_punctuation=True)
        >>> spacy_obj.process("Hello there. How are you? I'm fine, thanks.")
        ["Hello", "there", "How", "are", "you", "I", "'m", "fine", "thanks"]

        * Remove stopwords using default stopwords corpus of spacy from running text
        >>> spacy_obj = Spacy(stopwords_removal=True)
        >>> spacy_obj.process("The striped bats are hanging on their feet for the best")
        ["striped", "bats", "hanging", "feet", "best"]

        * Remove stopwords using default stopwords corpus of spacy + `new_stopwords` list from running text
        >>> spacy_obj = Spacy(stopwords_removal=True, new_stopwords=['bats', 'best'])
        >>> spacy_obj.process("The striped bats are hanging on their feet for the best")
        ["striped", "hanging", "feet"]

        * Remove stopwords using default stopwords corpus of spacy - `not_stopwords` list from running text
        >>> spacy_obj = Spacy(stopwords_removal=True, not_stopwords=['The', 'the', 'on'])
        >>> spacy_obj.process("The striped bats are hanging on their feet for the best")
        ["The", "striped", "bats", "hanging", "on", "feet", "the", "best"]

        * Replace URL with a normalized token `<URL>`
        >>> spacy_obj = Spacy(url_tagging=True)
        >>> spacy_obj.process("This is facebook http://facebook.com and github https://github.com")
        ['This', 'is', 'facebook', '<URL>', 'and', 'github', '<URL>']

        * Perform lemmatization on running text
        >>> spacy_obj = Spacy(lemmatization=True)
        >>> spacy_obj.process("The striped bats are hanging on their feet for best")
        ["The", "strip", "bat", "be", "hang", "on", "their", "foot", "for", "best"]

        * Perform NER on running text (NEs will be tagged with BIO tagging)
        >>> spacy_obj = Spacy(named_entity_recognition=True)
        >>> spacy_obj.process("Facebook was fined by Hewlett Packard for spending 100€")
        ["Facebook", "was", "fined", "by", "<Hewlett_ORG_B>", "<Packard_ORG_I>", "for", "spending",
        "<100_MONEY_B>", "<€_MONEY_I>"]

    Args:
        model: Spacy model that will be used to perform nlp operations. It will be downloaded if not present locally
        strip_multiple_whitespaces: If set to True, all multiple whitespaces will be reduced to only one white space
        remove_punctuation: If set to True, all punctuation from the running text will be removed
        stopwords_removal: If set to True, all stowpwords from the running text will be removed
        new_stopwords: List which contains custom defined stopwords that will be removed if `stopwords_removal=True`
        not_stopwords: List which contains custom defined stopwords that will not be considered as such, therefore won't
            be removed if `stopwords_removal=True`
        url_tagging: If set to True, all urls in the running text will be replaced with the `<URL>` token
        lemmatization: If set to True, each token in the running text will be brought to its lemma
        named_entity_recognition: If set to True, named entities recognized will be labeled in the form `<token_B_TAG>`
            or `<token_I_TAG>`, according to BIO tagging strategy
    """

    def __init__(self, model: str = 'en_core_web_sm', *,
                 strip_multiple_whitespaces: bool = True,
                 remove_punctuation: bool = False,
                 stopwords_removal: bool = False,
                 new_stopwords: List[str] = None,
                 not_stopwords: List[str] = None,
                 lemmatization: bool = False,
                 url_tagging: bool = False,
                 named_entity_recognition: bool = False):

        self.model = model
        self.stopwords_removal = stopwords_removal
        self.lemmatization = lemmatization
        self.strip_multiple_whitespaces = strip_multiple_whitespaces
        self.url_tagging = url_tagging
        self.remove_punctuation = remove_punctuation
        self.named_entity_recognition = named_entity_recognition

        # download the model if not present. In any case load it
        if model not in spacy.cli.info()['pipelines']:
            spacy.cli.download(model)
        self._nlp = spacy.load(model)

        # Adding custom rule of preserving '<URL>' token and in general token
        # wrapped by '<...>'
        prefixes = list(self._nlp.Defaults.prefixes)
        prefixes.remove('<')
        prefix_regex = spacy.util.compile_prefix_regex(prefixes)
        self._nlp.tokenizer.prefix_search = prefix_regex.search

        suffixes = list(self._nlp.Defaults.suffixes)
        suffixes.remove('>')
        suffix_regex = spacy.util.compile_suffix_regex(suffixes)
        self._nlp.tokenizer.suffix_search = suffix_regex.search

        self.not_stopwords_list = not_stopwords
        if not_stopwords is not None:
            for stopword in not_stopwords:
                self._nlp.vocab[stopword].is_stop = False

        self.new_stopwords_list = new_stopwords
        if new_stopwords is not None:
            for stopword in new_stopwords:
                self._nlp.vocab[stopword].is_stop = True

    def __tokenization_operation(self, text) -> List[Token]:
        """
        Splits the text in one-word tokens

        Args:
             text (str): Text to split in tokens

        Returns:
             List<str>: a list of words
        """
        pipe_to_disable = ['tagger', 'parser', 'textcat']

        if not self.lemmatization:
            pipe_to_disable.append('lemmatizer')
        if not self.named_entity_recognition:
            pipe_to_disable.append('ner')

        return list(self._nlp(text))

    def __stopwords_removal_operation(self, text) -> List[Token]:
        """
        Execute stopwords removal on input text with spacy

        Args:
            text (List[Token]):

        Returns:
            filtered_sentence (List<str>): list of words from the text, without the stopwords
        """
        filtered_sentence = [token for token in text if not token.is_stop]

        return filtered_sentence

    def __lemmatization_operation(self, text) -> List[Token]:
        """
        Execute lemmatization on input text with spacy

        Args:
            text (List[Token]):

        Returns:
            lemmatized_text (List<str>): List of the fords from the text, reduced to their lemmatized version
        """
        lemmas_to_tokenize = ' '.join([word.lemma_ for word in text])

        return self.__tokenization_operation(lemmas_to_tokenize)

    def __named_entity_recognition_operation(self, text) -> List[Token]:
        """
        Execute NER on input text with spacy

        Args:
            text List[str]: Text containing the entities

        Returns:
            word_entity: Dict of entity
        """
        labeled_entities = ' '.join([f"<{token.text}_{token.ent_type_}_{token.ent_iob_}>" if token.ent_type != 0
                                     else f"{token.text}" for token in text])

        return self.__tokenization_operation(labeled_entities)

    @staticmethod
    def __strip_multiple_whitespaces_operation(text) -> str:
        """
        Remove multiple whitespaces on input text

        Args:
            text (str):

        Returns:
            str: input text, multiple whitespaces removed
        """
        import re
        return re.sub(' +', ' ', text)

    def __url_tagging_operation(self, text) -> List[Token]:
        """
        Replaces urls with <URL> string on input text with spacy

        Args:
            text (list[Token]):

        Returns:
            text (list<str>): input text, <URL> instead of full urls
        """

        text_w_url_to_tokenize = ' '.join(["<URL>" if token.like_url else str(token) for token in text])

        return self.__tokenization_operation(text_w_url_to_tokenize)

    def __remove_punctuation(self, text) -> List[Token]:
        """
        Punctuation removal in spacy
        Args:
            text (list[Token]):
        Returns:
            string without punctuation
        """
        cleaned_text = [token for token in text if not token.is_punct]

        return cleaned_text

    @staticmethod
    def __token_to_string(token_field) -> List[str]:
        """
        After preprocessing with spacy the output was given as a list of str

        Args:
            token_field: List of tokens
        Returns:
            list of string
        """
        string_list = [token.text for token in token_field]

        return string_list

    def process(self, field_data: str) -> List[str]:
        """
        Args:
            field_data: content to be processed

        Returns:
            field_data: list of str or dict in case of named entity recognition

        """
        field_data = check_not_tokenized(field_data)
        if self.strip_multiple_whitespaces:
            field_data = self.__strip_multiple_whitespaces_operation(field_data)
        field_data = self.__tokenization_operation(field_data)
        if self.named_entity_recognition:
            field_data = self.__named_entity_recognition_operation(field_data)
        if self.remove_punctuation:
            field_data = self.__remove_punctuation(field_data)
        if self.stopwords_removal:
            field_data = self.__stopwords_removal_operation(field_data)
        if self.lemmatization:
            field_data = self.__lemmatization_operation(field_data)
        if self.url_tagging:
            field_data = self.__url_tagging_operation(field_data)

        return self.__token_to_string(field_data)

    def __eq__(self, other):
        if isinstance(other, Spacy):
            return self.model == other.model and \
                   self.strip_multiple_whitespaces == other.strip_multiple_whitespaces and \
                   self.remove_punctuation == other.remove_punctuation and \
                   self.stopwords_removal == other.stopwords_removal and \
                   self.new_stopwords_list == other.new_stopwords_list and \
                   self.not_stopwords_list == other.not_stopwords_list and \
                   self.lemmatization == other.lemmatization and \
                   self.url_tagging == other.url_tagging and \
                   self.named_entity_recognition == other.named_entity_recognition
        return False

    def __str__(self):
        return "Spacy"

    def __repr__(self):
        return f'Spacy(model={self.model}, strip_multiple_whitespace={self.strip_multiple_whitespaces}, ' \
               f'remove_punctuation={self.remove_punctuation}, stopwords_removal={self.stopwords_removal}, ' \
               f'new_stopwords={self.new_stopwords_list}, not_stopwords={self.not_stopwords_list}, ' \
               f'lemmatization={self.lemmatization}, url_tagging={self.url_tagging}, ' \
               f'named_entity_recognition={self.named_entity_recognition})'



import os.path
import shutil
import unittest
import yaml

from clayrs.content_analyzer import ItemAnalyzerConfig, JSONFile, ExogenousConfig, PropertiesFromDataset, \
    SkLearnTfIdf, NLTK, SentenceEmbeddingTechnique, Spacy, Sbert, WordEmbeddingTechnique, Gensim, FieldConfig, \
    ContentAnalyzer, Ekphrasis, Ratings, CSVFile, Rank
from clayrs.evaluation import EvalModel, Precision, PrecisionAtK, NDCG, MRR, FMeasureAtK
from clayrs.recsys import HoldOutPartitioning, NXPageRank, NXFullGraph, GraphBasedRS
from clayrs.utils.const import datasets_path
from clayrs.utils.report import Report

items_info = os.path.join(datasets_path, 'ml-100k', 'items_info.json')
rat_path = os.path.join(datasets_path, 'ml-100k', 'ratings.csv')


# Obviously it's impossible to test all methods, here we test a bunch
# but manual testing is required for the Report module
class TestReport(unittest.TestCase):

    @staticmethod
    def _build_ca_report():
        movies_ca_config = ItemAnalyzerConfig(
            source=JSONFile(items_info),
            id='movielens_id',
            output_directory='movies_codified/'
        )

        movies_ca_config.add_single_exogenous(ExogenousConfig(PropertiesFromDataset(field_name_list=['review'])))

        movies_ca_config.add_single_config(
            'plot',
            FieldConfig(content_technique=SkLearnTfIdf(),

                        preprocessing=[NLTK(stopwords_removal=True), Ekphrasis()])
        )

        movies_ca_config.add_multiple_config(
            'genres',
            [
                FieldConfig(content_technique=WordEmbeddingTechnique(Gensim()),
                            preprocessing=Spacy(stopwords_removal=True)),

                FieldConfig(content_technique=SentenceEmbeddingTechnique(Sbert()),
                            preprocessing=Spacy(stopwords_removal=True, remove_punctuation=True)),
            ]
        )

        c_analyzer = ContentAnalyzer(movies_ca_config)

        Report(output_dir='test_report').yaml(content_analyzer=c_analyzer)

    def test_ca_yaml(self):

        self._build_ca_report()

        self.assertTrue(os.path.isfile('test_report/ca_report.yml'))

        with open('test_report/ca_report.yml', "r") as stream:
            result = yaml.safe_load(stream)

        # Check that exogenous representation codified is present
        exogenous_representations_dict = result.get('exogenous_representations')
        self.assertIsNotNone(exogenous_representations_dict)
        self.assertIsNotNone(exogenous_representations_dict.get('PropertiesFromDataset'))

        # Check that field representations codified are present
        field_dict = result.get("field_representations")
        self.assertIsNotNone(field_dict)

        # Representations for plot field
        plot_dict = field_dict.get('plot_0')
        self.assertIsNotNone(plot_dict)

        self.assertIsNotNone(plot_dict.get('SkLearnTfIdf'))

        plot_preprocessing_dict = plot_dict.get('preprocessing')
        self.assertIsNotNone(plot_preprocessing_dict)
        self.assertIsNotNone(plot_preprocessing_dict.get('NLTK'))
        self.assertIsNotNone(plot_preprocessing_dict.get('Ekphrasis'))

        # Representation 0 for genres field
        genres_0_dict = field_dict.get('genres_0')
        self.assertIsNotNone(genres_0_dict)

        self.assertIsNotNone(genres_0_dict.get('WordEmbeddingTechnique'))

        genres_0_preprocessing_dict = genres_0_dict.get('preprocessing')
        self.assertIsNotNone(genres_0_preprocessing_dict)
        self.assertIsNotNone(genres_0_preprocessing_dict.get('Spacy'))

        # Representation 1 for genres field
        genres_1_dict = field_dict.get('genres_1')
        self.assertIsNotNone(genres_1_dict)

        self.assertIsNotNone(genres_1_dict.get('SentenceEmbeddingTechnique'))

        genres_1_preprocessing_dict = genres_1_dict.get('preprocessing')
        self.assertIsNotNone(genres_1_preprocessing_dict)
        self.assertIsNotNone(genres_1_preprocessing_dict.get('Spacy'))

    @staticmethod
    def _build_rs_report():
        original_rat = Ratings(CSVFile(rat_path))

        pt = HoldOutPartitioning()

        train_list, test_list = pt.split_all(original_rat)

        alg = NXPageRank()

        graph = NXFullGraph(train_list[0])

        gbrs = GraphBasedRS(alg, graph=graph)
        gbrs.rank(test_list[0])

        Report(output_dir='test_report/only_ratings').yaml(original_ratings=original_rat)
        Report(output_dir='test_report/only_partitioning').yaml(partitioning_technique=pt)
        Report(output_dir='test_report/only_recsys').yaml(recsys=gbrs)
        Report(output_dir='test_report/all_rs_module').yaml(original_ratings=original_rat,
                                                            partitioning_technique=pt,
                                                            recsys=gbrs)

    def test_rs_yaml(self):
        self._build_rs_report()

        self.assertTrue(os.path.isfile('test_report/only_ratings/rs_report.yml'))
        self.assertTrue(os.path.isfile('test_report/only_partitioning/rs_report.yml'))
        self.assertTrue(os.path.isfile('test_report/only_recsys/rs_report.yml'))
        self.assertTrue(os.path.isfile('test_report/all_rs_module/rs_report.yml'))

        # test only ratings report
        with open('test_report/only_ratings/rs_report.yml', "r") as stream:
            result = yaml.safe_load(stream)

        self.assertTrue(len(result) == 1)
        self.assertIsNotNone(result.get('interactions'))

        # test only partitioning report
        with open('test_report/only_partitioning/rs_report.yml', "r") as stream:
            result = yaml.safe_load(stream)

        self.assertTrue(len(result) == 1)
        partitioning_dict = result.get('partitioning')
        self.assertIsNotNone(partitioning_dict)
        self.assertIsNotNone(partitioning_dict.get('HoldOutPartitioning'))

        # test only recsys report
        with open('test_report/only_recsys/rs_report.yml', "r") as stream:
            result = yaml.safe_load(stream)

        self.assertTrue(len(result) == 1)
        recsys_dict = result.get('recsys')
        self.assertIsNotNone(recsys_dict)
        self.assertIsNotNone(recsys_dict.get('GraphBasedRS'))

        # test all rs module report
        with open('test_report/all_rs_module/rs_report.yml', "r") as stream:
            result = yaml.safe_load(stream)

        self.assertTrue(len(result) == 3)

        self.assertIsNotNone(result.get('interactions'))

        partitioning_dict = result.get('partitioning')
        self.assertIsNotNone(partitioning_dict)
        self.assertIsNotNone(partitioning_dict.get('HoldOutPartitioning'))

        recsys_dict = result.get('recsys')
        self.assertIsNotNone(recsys_dict)
        self.assertIsNotNone(recsys_dict.get('GraphBasedRS'))

    def test_rs_yaml_error(self):
        original_rat = Ratings(CSVFile(rat_path))

        pt = HoldOutPartitioning()

        train_list, test_list = pt.split_all(original_rat)

        alg = NXPageRank()

        graph = NXFullGraph(train_list[0])

        gbrs = GraphBasedRS(alg, graph=graph)

        # try to build report without calling the rank method first
        with self.assertRaises(ValueError):
            Report(output_dir="this doesn't work").yaml(recsys=gbrs)

    @staticmethod
    def _build_eva_report():

        # rank same as truth since we don't care about results
        rank = Rank(CSVFile(rat_path))
        truth = Ratings(CSVFile(rat_path))

        em = EvalModel([rank],
                       [truth],
                       metric_list=[Precision(),
                                    PrecisionAtK(k=2),
                                    NDCG(),
                                    MRR(),
                                    FMeasureAtK(k=1)])
        em.fit()

        Report('test_report').yaml(eval_model=em)

    def test_eva_yaml(self):
        self._build_eva_report()

        self.assertTrue(os.path.isfile('test_report/eva_report.yml'))

        with open('test_report/eva_report.yml', "r") as stream:
            result = yaml.safe_load(stream)

        self.assertIsNotNone(result.get('n_split'))
        self.assertTrue(result.get('n_split') == 1)

        metrics_dict = result.get('metrics')
        self.assertIsNotNone(metrics_dict)
        self.assertIsNotNone(metrics_dict.get('Precision'))
        self.assertIsNotNone(metrics_dict.get('PrecisionAtK'))
        self.assertIsNotNone(metrics_dict.get('NDCG'))
        self.assertIsNotNone(metrics_dict.get('MRR'))
        self.assertIsNotNone(metrics_dict.get('FMeasureAtK'))

        sys_results_dict = result.get('sys_results')
        self.assertIsNotNone(sys_results_dict.get('sys - fold1'))
        self.assertIsNotNone(sys_results_dict.get('sys - mean'))

        sys_fold_result = sys_results_dict.get('sys - fold1')
        self.assertIsNotNone(sys_fold_result.get('Precision - macro'))
        self.assertIsNotNone(sys_fold_result.get('Precision@2 - macro'))
        self.assertIsNotNone(sys_fold_result.get('NDCG'))
        self.assertIsNotNone(sys_fold_result.get('MRR'))
        self.assertIsNotNone(sys_fold_result.get('F1@1 - macro'))

    def test_eva_yaml_error(self):
        # rank same as truth since we don't care about results
        rank = Rank(CSVFile(rat_path))
        truth = Ratings(CSVFile(rat_path))

        em = EvalModel([rank],
                       [truth],
                       metric_list=[Precision(),
                                    PrecisionAtK(k=2),
                                    NDCG(),
                                    MRR(),
                                    FMeasureAtK(k=1)])

        # try to build report without calling the fit method first
        with self.assertRaises(ValueError):
            Report('test_report').yaml(eval_model=em)

    @classmethod
    def tearDownClass(cls) -> None:
        shutil.rmtree('test_report')


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import TestCase

from clayrs.content_analyzer import Ratings
from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsDict
from clayrs.recsys.content_based_algorithm.exceptions import NoRatedItems, EmptyUserRatings
from clayrs.recsys.content_based_algorithm.regressor.linear_predictor import LinearPredictor

from clayrs.recsys.content_based_algorithm.regressor.regressors import SkLinearRegression, \
    SkSGDRegressor, SkARDRegression, SkHuberRegressor, SkPassiveAggressiveRegressor, SkBayesianRidge, SkRidge, \
    Regressor
import os
import pandas as pd

from test import dir_test_files


def for_each_model(test_func):
    def wrapper(self, *args, **kwargs):
        for model in self.models_list:
            with self.subTest(current_model=model):
                test_func(*((self, model) + args), **kwargs)

    return wrapper


class TestRegression(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        ratings = pd.DataFrame.from_records([
            ("A000", "tt0112281", 3.5, "54654675"),
            ("A000", "tt0112302", 3.5, "54654675"),
            ("A001", "tt0114576", 4, "54654675"),
            ("A001", "tt0112896", 2, "54654675"),
            ("A000", "tt0112346", 2, "54654675"),
            ("A000", "tt0112453", 1, "54654675"),
            ("A002", "tt0112453", 1, "54654675"),
            ("A002", "tt0113497", 3.5, "54654675"),
            ("A003", "tt0112453", 1, "54654675")],
            columns=["user_id", "item_id", "score", "timestamp"])
        cls.ratings = Ratings.from_dataframe(ratings)

        cls.filter_list = ['tt0112641', 'tt0112760', 'tt0112896', 'tt0113497']

        movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')

        cls.available_loaded_items = LoadedContentsDict(movies_dir)

        # IMPORTANT! All models to test. If another model is implemented, append it to this list
        cls.models_list = [SkLinearRegression(), SkRidge(), SkBayesianRidge(),
                           SkSGDRegressor(), SkARDRegression(), SkHuberRegressor(),
                           SkPassiveAggressiveRegressor()]

    @for_each_model
    def test_predict_single_representation(self, model: Regressor):
        lm = model

        # Single representation
        alg = LinearPredictor({'Plot': ['tfidf']}, lm)

        user_ratings = self.ratings.get_user_interactions("A000")

        alg.process_rated(user_ratings, self.available_loaded_items)
        alg.fit()

        # predict with filter_list
        res_filtered = alg.predict(user_ratings, self.available_loaded_items, filter_list=self.filter_list)
        item_scored_set = set([interaction_filtered.item_id for interaction_filtered in res_filtered])
        self.assertEqual(len(item_scored_set), len(self.filter_list))
        self.assertCountEqual(item_scored_set, self.filter_list)

        # predict without filter_list
        res_all_unrated = alg.predict(user_ratings, self.available_loaded_items)
        item_rated_set = set([interaction.item_id for interaction in user_ratings])
        item_scored_set = set([interaction_all.item_id for interaction_all in res_all_unrated])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_scored = item_scored_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_scored), 0)

    @for_each_model
    def test_predict_multiple_representations(self, model: Regressor):
        lm = model

        # Multiple representations filtered only items with score >= 2
        alg = LinearPredictor({'Plot': ['tfidf', 'embedding'],
                               'Genre': ['tfidf', 'embedding'],
                               'imdbRating': [0]}, lm, only_greater_eq=2)

        user_ratings = self.ratings.get_user_interactions("A000")

        alg.process_rated(user_ratings, self.available_loaded_items)
        alg.fit()

        # predict with filter_list
        res_filtered = alg.predict(user_ratings, self.available_loaded_items, filter_list=self.filter_list)
        item_scored_set = set([interaction_filtered.item_id for interaction_filtered in res_filtered])
        self.assertEqual(len(item_scored_set), len(self.filter_list))
        self.assertCountEqual(item_scored_set, self.filter_list)

        # predict without filter_list
        res_all_unrated = alg.predict(user_ratings, self.available_loaded_items)
        item_rated_set = set([interaction.item_id for interaction in user_ratings])
        item_scored_set = set([interaction_all.item_id for interaction_all in res_all_unrated])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_scored = item_scored_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_scored), 0)

    @for_each_model
    def test_rank_single_representation(self, model: Regressor):
        lm = model

        # Single representation
        alg = LinearPredictor({'Plot': ['tfidf']}, lm)

        user_ratings = self.ratings.get_user_interactions("A000")

        alg.process_rated(user_ratings, self.available_loaded_items)
        alg.fit()

        # rank with filter_list
        res_filtered = alg.rank(user_ratings, self.available_loaded_items, filter_list=self.filter_list)
        item_ranked_set = set([interaction_filtered.item_id for interaction_filtered in res_filtered])
        self.assertEqual(len(item_ranked_set), len(self.filter_list))
        self.assertCountEqual(item_ranked_set, self.filter_list)

        # rank without filter_list
        res_all_unrated = alg.rank(user_ratings, self.available_loaded_items)
        item_rated_set = set([interaction.item_id for interaction in user_ratings])
        item_ranked_set = set([interaction_all.item_id for interaction_all in res_all_unrated])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_ranked = item_ranked_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_ranked), 0)

        # rank with n_recs specified
        n_recs = 5
        res_n_recs = alg.rank(user_ratings, self.available_loaded_items, n_recs)
        self.assertEqual(len(res_n_recs), n_recs)
        item_rated_set = set([interaction.item_id for interaction in user_ratings])
        item_ranked_set = set([interaction_nrecs.item_id for interaction_nrecs in res_n_recs])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_ranked = item_ranked_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_ranked), 0)

    @for_each_model
    def test_rank_multiple_representations(self, model: Regressor):
        lm = model

        # Multiple representations filtered only items with score >= 2
        alg = LinearPredictor({'Plot': ['tfidf', 'embedding'],
                               'Genre': ['tfidf', 'embedding'],
                               'imdbRating': [0]}, lm, only_greater_eq=2)

        user_ratings = self.ratings.get_user_interactions("A000")

        alg.process_rated(user_ratings, self.available_loaded_items)
        alg.fit()

        # rank with filter_list
        res_filtered = alg.rank(user_ratings, self.available_loaded_items, filter_list=self.filter_list)
        item_ranked_set = set([interaction_filtered.item_id for interaction_filtered in res_filtered])
        self.assertEqual(len(item_ranked_set), len(self.filter_list))
        self.assertCountEqual(item_ranked_set, self.filter_list)

        # rank without filter_list
        res_all_unrated = alg.rank(user_ratings, self.available_loaded_items)
        item_rated_set = set([interaction.item_id for interaction in user_ratings])
        item_ranked_set = set([interaction_all.item_id for interaction_all in res_all_unrated])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_ranked = item_ranked_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_ranked), 0)

        # rank with n_recs specified
        n_recs = 5
        res_n_recs = alg.rank(user_ratings, self.available_loaded_items, n_recs)
        self.assertEqual(len(res_n_recs), n_recs)
        item_rated_set = set([interaction.item_id for interaction in user_ratings])
        item_ranked_set = set([interaction_nrecs.item_id for interaction_nrecs in res_all_unrated])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_ranked = item_ranked_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_ranked), 0)

    def test_raise_errors(self):
        # No Item available locally
        ratings = pd.DataFrame.from_records([
            ("A000", "non existent", 1, "54654675")],
            columns=["user_id", "item_id", "score", "timestamp"])
        ratings = Ratings.from_dataframe(ratings)

        alg = LinearPredictor({'Plot': ['tfidf']}, SkLinearRegression())
        user_ratings = ratings.get_user_interactions('A000')

        with self.assertRaises(NoRatedItems):
            alg.process_rated(user_ratings, self.available_loaded_items)

        # User has no ratings
        user_ratings = []

        alg = LinearPredictor({'Plot': ['tfidf']}, SkLinearRegression())

        with self.assertRaises(EmptyUserRatings):
            alg.process_rated(user_ratings, self.available_loaded_items)

        with self.assertRaises(EmptyUserRatings):
            alg.rank(user_ratings, self.available_loaded_items)


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase
from clayrs.recsys.content_based_algorithm.centroid_vector.similarities import CosineSimilarity
import numpy as np


class TestCosineSimilarity(TestCase):
    def test_perform(self):
        sim = CosineSimilarity()

        a = np.array([5, 9, 7, 8, 3, 5, 4, 2, 6, 4])
        b = np.array([8, 1, 3, 10, 8, 4, 9, 2, 1, 6])
        self.assertAlmostEqual(sim.perform(a, b), 0.7552110293516224)

        a = np.array([0, 0, 0])
        b = np.array([1, 1, 1])
        self.assertEqual(sim.perform(a, b), 0)

        a = np.array([1, 1, 1])
        b = np.array([0, 0, 0])
        self.assertEqual(sim.perform(a, b), 0)

        a = np.array([0, 0, 0])
        b = np.array([0, 0, 0])
        self.assertEqual(sim.perform(a, b), 0)

        a = np.array([1, 1, 1])
        b = np.array([1, 1, 1])
        self.assertEqual(sim.perform(a, b), 1)

        a = np.array([1, 1, 1])
        b = np.array([-1, -1, -1])
        self.assertEqual(sim.perform(a, b), -1)



from __future__ import annotations
from collections import defaultdict

from typing import Set, List, Tuple, TYPE_CHECKING

import abc
from abc import ABC

from sklearn.model_selection import KFold, train_test_split
from sklearn.utils import resample

if TYPE_CHECKING:
    from clayrs.content_analyzer.ratings_manager.ratings import Interaction

from clayrs.content_analyzer.ratings_manager.ratings import Ratings
from clayrs.utils.const import logger
from clayrs.utils.context_managers import get_progbar


class Partitioning(ABC):
    """
    Abstract class for partitioning technique. Each class must implement the `split_single()` method which specify how
    data for a single user will be split

    Args:
        skip_user_error:
            If set to True, users for which data can't be split will be skipped and only a warning will be logged when
            calling the `split_all()` method. Otherwise, a `ValueError` exception is raised
    """

    def __init__(self, skip_user_error: bool = True):
        self.__skip_user_error = skip_user_error

    @property
    def skip_user_error(self):
        return self.__skip_user_error

    @abc.abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abc.abstractmethod
    def split_single(self, user_ratings: List[Interaction]) -> Tuple[List[List[Interaction]], List[List[Interaction]]]:
        """
        Abstract method in which each partitioning technique must specify how to split data for a single user

        Args:
            user_ratings: List of `Interaction` objects of a single user

        Returns:
            Two lists, where the first contains one list of `Interaction` objects for each split that will
                constitute the *train set* of the user, the second contains one list of `Interaction` objects for each split
                that will constitute the *test set* for the user
        """
        raise NotImplementedError

    def split_all(self, ratings_to_split: Ratings, user_id_list: Set[str] = None) -> Tuple[List[Ratings],
                                                                                           List[Ratings]]:
        """
        Concrete method that splits, for every user in the `ratings_to_split` parameter, the original ratings
        into *train set* and *test set*.
        If a `user_id_list` parameter is set, the method will do the splitting only for the users
        specified inside the list.

        The method returns two lists:

        * The first contains all train set for each split (if the partitioning technique returns more than one split
        e.g. KFold)
        * The second contains all test set for each split (if the partitioning technique returns more than one split
        e.g. KFold)

        Obviously the two lists will have the same length, and to the *train set* in position $i$ corresponds the
        *truth set* at position $i$

        Args:
            ratings_to_split: `Ratings` object which contains the interactions of the users that must be splitted
                into *train set* and *test set*
            user_id_list: The set of users for which splitting will be done. If set, splitting will be performed only
                for users inside the list. Otherwise, splitting will be performed for all users in `ratings_to_split`
                parameter

        Raises:
            ValueError: if `skip_user_error=True` in the constructor and for some users splitting can't be performed
        """

        if user_id_list is None:
            user_id_list = set(ratings_to_split.user_id_column)

        # {
        #   0: {'train': {'u1': u1_interactions_train0, 'u2': u2_interactions_train0}},
        #       'test': {'u1': u1_interactions_test0, 'u2': u2_interactions_test0}},
        #
        #   1: {'train': {'u1': u1_interactions_train1, 'u2': u2_interactions_train1}},
        #       'test': {'u1': u1_interactions_test1, 'u2': u2_interactions_test1}
        #  }
        train_test_dict = defaultdict(lambda: defaultdict(list))

        with get_progbar(user_id_list) as pbar:

            pbar.set_description("Performing {}".format(str(self)))
            for user_id in pbar:
                user_ratings = ratings_to_split.get_user_interactions(user_id)
                try:
                    user_train_list, user_test_list = self.split_single(user_ratings)
                    for split_number, (single_train, single_test) in enumerate(zip(user_train_list, user_test_list)):
                        # we set for each split the train_set and test_set of every user u1
                        # eg.
                        #     train_test_dict[0]['train']['u1'] = u1_interactions_train0
                        #     train_test_dict[0]['test']['u1'] = u1_interactions_test0
                        # train_test_dict[split_number]['train'][user_id] = single_train
                        # train_test_dict[split_number]['test'][user_id] = single_test
                        train_test_dict[split_number]['train'].extend(single_train)
                        train_test_dict[split_number]['test'].extend(single_test)

                except ValueError as e:
                    if self.skip_user_error:
                        logger.warning(str(e) + "\nThe user {} will be skipped".format(user_id))
                        continue
                    else:
                        raise e

        train_list = [Ratings.from_list(train_test_dict[split]['train'])
                      for split in train_test_dict]

        test_list = [Ratings.from_list(train_test_dict[split]['test'])
                     for split in train_test_dict]

        return train_list, test_list


class KFoldPartitioning(Partitioning):
    """
    Class that performs K-Fold partitioning

    Args:
        n_splits (int): Number of splits. Must be at least 2
        shuffle:
            Whether to shuffle the data before splitting into batches.
            Note that the samples within each split will not be shuffled.
        random_state:
            When `shuffle` is True, `random_state` affects the ordering of the
            indices, which controls the randomness of each fold. Otherwise, this
            parameter has no effect.
            Pass an int for reproducible output across multiple function calls.
        skip_user_error:
            If set to True, users for which data can't be split will be skipped and only a warning will be logged when
            calling the `split_all()` method. Otherwise, a `ValueError` exception is raised
    """

    def __init__(self, n_splits: int = 2, shuffle: bool = True, random_state: int = None,
                 skip_user_error: bool = True):
        self.__kf = KFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state)

        super(KFoldPartitioning, self).__init__(skip_user_error)

    def split_single(self, user_ratings: List[Interaction]) -> Tuple[List[List[Interaction]], List[List[Interaction]]]:
        """
        Method which splits in $k$ splits both in *train set* and *test set* the ratings of a single user

        Args:
            user_ratings: List of `Interaction` objects of a single user

        Returns:
            Two lists, where the first contains one list of `Interaction` objects for each split that will
                constitute the *train set* of the user, the second contains one list of `Interaction` objects for each split
                that will constitute the *test set* for the user
        """
        split_result = self.__kf.split(user_ratings)

        user_train_list = []
        user_test_list = []
        # split_result contains index of the ratings which must constitutes train set and test set
        for train_set_indexes, test_set_indexes in split_result:
            user_interactions_train = [user_ratings[index] for index in train_set_indexes]

            user_interactions_test = [user_ratings[index] for index in test_set_indexes]

            user_train_list.append(user_interactions_train)
            user_test_list.append(user_interactions_test)

        return user_train_list, user_test_list

    def __str__(self):
        return "KFoldPartitioning"

    def __repr__(self):
        return f"KFoldPartitioning(n_splits={self.__kf.n_splits}, shuffle={self.__kf.shuffle}, " \
               f"random_state={self.__kf.random_state}, skip_user_error={self.skip_user_error})"


class HoldOutPartitioning(Partitioning):
    """
    Class that performs Hold-Out partitioning

    Args:
        train_set_size: Should be between 0.0 and 1.0 and represent the proportion of the ratings to
            ***hold*** in the train set for each user.
        random_state:
            Controls the shuffling applied to the data before applying the split.
            Pass an int for reproducible output across multiple function calls.
        shuffle:
            Whether or not to shuffle the data before splitting.
        skip_user_error:
            If set to True, users for which data can't be split will be skipped and only a warning will be logged when
            calling the `split_all()` method. Otherwise, a `ValueError` exception is raised
    """

    def __init__(self, train_set_size: float = 0.8, shuffle: bool = True, random_state: int = None,
                 skip_user_error: bool = True):
        self._check_percentage(train_set_size)
        self.__train_set_size = train_set_size
        self.__test_set_size = (1 - train_set_size)
        self.__random_state = random_state
        self.__shuffle = shuffle

        super().__init__(skip_user_error)

    @staticmethod
    def _check_percentage(percentage: float):
        if (percentage <= 0) or (percentage >= 1):
            raise ValueError("The train set size must be a float in the (0, 1) interval")

    def split_single(self, user_ratings: List[Interaction]) -> Tuple[List[List[Interaction]], List[List[Interaction]]]:
        """
        Method which splits *train set* and *test set* the ratings of a single user by holding in the train
        set the percentage of data specified in `train_set_size` in the constructor

        Args:
            user_ratings: List of `Interaction` objects of a single user

        Returns:
            Two lists, where the first contains one list of `Interaction` objects that will
                constitute the *train set* of the user, the second contains one list of `Interaction` objects
                that will constitute the *test set* for the user
        """
        interactions_train, interactions_test = train_test_split(user_ratings,
                                                                 train_size=self.__train_set_size,
                                                                 test_size=self.__test_set_size,
                                                                 shuffle=self.__shuffle,
                                                                 random_state=self.__random_state)

        user_train_list = [interactions_train]
        user_test_list = [interactions_test]

        return user_train_list, user_test_list

    def __str__(self):
        return "HoldOutPartitioning"

    def __repr__(self):
        return f"HoldOutPartitioning(train_set_size={self.__train_set_size}, shuffle={self.__shuffle}, " \
               f"random_state={self.__random_state}, skip_user_error={self.skip_user_error})"


class BootstrapPartitioning(Partitioning):
    """
    Class that performs Bootstrap Partitioning.

    The bootstrap partitioning consists in executing $n$ extractions with replacement for each user from the original
    interaction frame, where $n$ is the length of the user interactions:

        * The sampled data will be part of the ***train set***
        * All the data which is part of the original dataset but was not sampled will be part of the ***test set***

    The bootstrap partitioning can **change** the original data distribution, since during the extraction phase you
    could sample the same data more than once

    Args:
        random_state:
            Controls the shuffling applied to the data before applying the split.
            Pass an int for reproducible output across multiple function calls.
        skip_user_error:
            If set to True, users for which data can't be split will be skipped and only a warning will be logged when
            calling the `split_all()` method. Otherwise, a `ValueError` exception is raised
    """

    def __init__(self, random_state: int = None, skip_user_error: bool = True):
        super().__init__(skip_user_error)

        self.__random_state = random_state

    def split_single(self, user_ratings: List[Interaction]) -> Tuple[List[List[Interaction]], List[List[Interaction]]]:
        """
        Method which splits *train set* and *test set* the ratings of a single user by performing $n$ extraction with
        replacement of the user interactions, where $n$ is the number of its interactions.
        The interactions which are not sampled will be part of the *test set*

        Args:
            user_ratings: List of `Interaction` objects of a single user

        Returns:
            Two lists, where the first contains one list of `Interaction` objects that will
                constitute the *train set* of the user, the second contains one list of `Interaction` objects
                that will constitute the *test set* for the user
        """

        interactions_train = resample(user_ratings,
                                      replace=True,
                                      n_samples=len(user_ratings),
                                      random_state=self.__random_state)

        interactions_test = [interaction for interaction in user_ratings
                             if interaction not in interactions_train]

        user_train_list = [interactions_train]
        user_test_list = [interactions_test]

        if len(interactions_test) == 0:
            raise ValueError("The test set for the user is empty! Try increasing the number of its interactions!")

        return user_train_list, user_test_list

    def __str__(self):
        return "BootstrapPartitioning"

    def __repr__(self):
        return f"BootstrapPartitioning(random_state={self.__random_state}, skip_user_error={self.skip_user_error})"



import pandas as pd
from unittest import TestCase

from clayrs.content_analyzer import Ratings
from clayrs.recsys.content_based_algorithm.exceptions import NotPredictionAlg
from clayrs.recsys.graph_based_algorithm.page_rank.nx_page_rank import NXPageRank
from clayrs.recsys.graphs import NXFullGraph

ratings = pd.DataFrame.from_records([
    ("A000", "tt0114576", 1, "54654675"),
    ("A000", "tt0112453", 0.5, "54654675"),
    ("A001", "tt0114576", 0.5, "54654675"),
    ("A001", "tt0112896", 0, "54654675"),
    ("A000", "tt0113041", 0.75, "54654675"),
    ("A002", "tt0112453", 0.5, "54654675"),
    ("A002", "tt0113497", 0.5, "54654675"),
    ("A003", "tt0112453", 0, "54654675")],
    columns=["from_id", "to_id", "score", "timestamp"])
ratings = Ratings.from_dataframe(ratings, timestamp_column="timestamp")

test_set = pd.DataFrame.from_records([
    ("A000", "tt0113497", 1, "54654675"),
    ("A001", "tt0112453", 0.5, "54654675"),
    ("A002", "tt0114576", 0.5, "54654675"),
    ("A003", "tt0114576", 0, "54654675")],
    columns=["from_id", "to_id", "score", "timestamp"])
test_set = Ratings.from_dataframe(test_set, timestamp_column="timestamp")


class TestNXPageRank(TestCase):

    def setUp(self) -> None:
        self.graph = NXFullGraph(ratings)

    def test_predict(self):
        alg = NXPageRank()

        # Will raise Exception since it's not a Score Prediction Algorithm
        with self.assertRaises(NotPredictionAlg):
            alg.predict({'A000'}, self.graph, test_set)

    def test_rank(self):
        # test not personalized
        alg = NXPageRank()

        # rank with None methodology (all unrated items will be ranked)
        res_all_unrated = alg.rank({'A000'}, self.graph, test_set, methodology=None, num_cpus=1)
        item_rated_set = set([interaction.item_id for interaction in ratings.get_user_interactions("A000")])
        item_ranked_set = set([pred_interaction.item_id for pred_interaction in res_all_unrated])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_ranked = item_ranked_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_ranked), 0)

        # rank with n_recs specified
        n_recs = 1
        res_n_recs = alg.rank({'A000'}, self.graph, test_set, n_recs)
        self.assertEqual(len(res_n_recs), n_recs)
        item_rated_set = set([interaction.item_id for interaction in ratings.get_user_interactions("A000")])
        item_ranked_set = set([pred_interaction.item_id for pred_interaction in res_n_recs])
        # We expect this to be empty, since the alg should rank only unrated items (unless in filter list)
        rated_in_ranked = item_ranked_set.intersection(item_rated_set)
        self.assertEqual(len(rated_in_ranked), 0)

        # test personalized
        alg = NXPageRank(personalized=True)
        result_personalized = alg.rank({'A000'}, self.graph, test_set, num_cpus=1)

        alg = NXPageRank()
        result_not_personalized = alg.rank({'A000'}, self.graph, test_set, num_cpus=1)

        self.assertNotEqual(result_personalized, result_not_personalized)

        # test with custom parameters
        alg = NXPageRank(alpha=0.9, max_iter=4, tol=1e-3)
        result_custom = alg.rank({'A000'}, self.graph, test_set, num_cpus=1)

        alg = NXPageRank()
        result_not_custom = alg.rank({'A000'}, self.graph, test_set, num_cpus=1)

        self.assertNotEqual(result_custom, result_not_custom)



import os
from clayrs.utils.custom_logger import get_custom_logger

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
root_path = os.path.join(THIS_DIR, '../../')
contents_path = os.path.join(root_path, 'contents/')
datasets_path = os.path.join(root_path, 'datasets/')

logger = get_custom_logger('custom_logger')



import os
from unittest import TestCase
import pandas as pd

from clayrs.content_analyzer import Ratings
from clayrs.recsys import LinearPredictor, SkLinearRegression, TrainingItemsMethodology
from clayrs.recsys.content_based_algorithm.classifier.classifiers import SkSVC
from clayrs.recsys.recsys import GraphBasedRS, ContentBasedRS
from clayrs.recsys.content_based_algorithm.classifier.classifier_recommender import ClassifierRecommender
from clayrs.recsys.content_based_algorithm.exceptions import NotPredictionAlg, NotFittedAlg
from clayrs.recsys.graph_based_algorithm.page_rank.nx_page_rank import NXPageRank
from clayrs.recsys.graphs import NXFullGraph

from test import dir_test_files

train_ratings = pd.DataFrame.from_records([
    ("A000", "tt0114576", 5, "54654675"),
    ("A001", "tt0114576", 3, "54654675"),
    ("A001", "tt0112896", 1, "54654675"),
    ("A000", "tt0113041", 1, "54654675"),
    ("A002", "tt0112453", 2, "54654675"),
    ("A002", "tt0113497", 4, "54654675"),
    ("A003", "tt0112453", 1, "54654675"),
    ("A003", "tt0113497", 4, "54654675")],
    columns=["from_id", "to_id", "score", "timestamp"])
train_ratings = Ratings.from_dataframe(train_ratings)

# No locally available items for A000
train_ratings_some_missing = pd.DataFrame.from_records([
    ("A000", "not_existent1", 5, "54654675"),
    ("A001", "tt0114576", 3, "54654675"),
    ("A001", "tt0112896", 1, "54654675"),
    ("A000", "not_existent2", 5, "54654675")],
    columns=["from_id", "to_id", "score", "timestamp"])
train_ratings_some_missing = Ratings.from_dataframe(train_ratings_some_missing)

test_ratings = pd.DataFrame.from_records([
    ("A000", "tt0114388", None),
    ("A000", "tt0112302", None),
    ("A001", "tt0113189", None),
    ("A001", "tt0113228", None),
    ("A002", "tt0114319", None),
    ("A002", "tt0114709", None),
    ("A003", "tt0114885", None)],
    columns=["from_id", "to_id", "score"])
test_ratings = Ratings.from_dataframe(test_ratings)


# Each of the cbrs algorithm has its own class tests, so we just take
# one cbrs alg as example. The behaviour is the same for all cbrs alg
class TestContentBasedRS(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.movies_multiple = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')

    def test_fit(self):
        # Test fit with cbrs algorithm
        alg = LinearPredictor({'Plot': ['tfidf', 'embedding']}, SkLinearRegression())
        cbrs = ContentBasedRS(alg, train_ratings, self.movies_multiple)

        cbrs.fit(num_cpus=1)

        # For the following user the algorithm could be fit
        self.assertIsNotNone(cbrs._user_fit_dic.get("A000"))
        self.assertIsNotNone(cbrs._user_fit_dic.get("A001"))
        self.assertIsNotNone(cbrs._user_fit_dic.get("A002"))
        self.assertIsNotNone(cbrs._user_fit_dic.get("A003"))

        # Test fit with the cbrs algorithm
        # For user A000 no items available locally, so the alg will not be fit for it
        cbrs_missing = ContentBasedRS(alg, train_ratings_some_missing, self.movies_multiple)

        cbrs_missing.fit(num_cpus=1)

        # For user A000 the alg could not be fit, but it could for A001
        self.assertIsNone(cbrs_missing._user_fit_dic.get("A000"))
        self.assertIsNotNone(cbrs_missing._user_fit_dic.get("A001"))

    def test_raise_error_without_fit(self):
        alg = LinearPredictor({'Plot': ['tfidf', 'embedding']}, SkLinearRegression())
        cbrs = ContentBasedRS(alg, train_ratings, self.movies_multiple)

        with self.assertRaises(NotFittedAlg):
            cbrs.rank(train_ratings, num_cpus=1)

        with self.assertRaises(NotFittedAlg):
            cbrs.predict(train_ratings, num_cpus=1)

    def test_rank(self):
        # Test fit with the cbrs algorithm
        alg = LinearPredictor({'Plot': ['tfidf', 'embedding']}, SkLinearRegression())
        cbrs = ContentBasedRS(alg, train_ratings, self.movies_multiple)

        # we must fit the algorithm in order to rank
        cbrs.fit(num_cpus=1)

        # Test ranking with the cbrs algorithm on specified items
        result_rank_filtered = cbrs.rank(test_ratings, num_cpus=1)
        self.assertEqual(len(result_rank_filtered), len(test_ratings))

        # Test ranking with the cbrs algorithm on all available unseen items
        result_rank_all = cbrs.rank(test_ratings, methodology=None, num_cpus=1)
        self.assertTrue(len(result_rank_all) != 0)

        # Test top-n ranking with the cbrs algorithm for only some users
        result_rank_numbered = cbrs.rank(test_ratings, n_recs=2, methodology=None, user_id_list=["A000", "A003"],
                                         num_cpus=1)
        self.assertEqual(set(result_rank_numbered.user_id_column), {"A000", "A003"})
        for user in {"A000", "A003"}:
            result_single = result_rank_numbered.get_user_interactions(user)
            self.assertTrue(len(result_single) == 2)

        # Test ranking with alternative methodology
        result_different_meth = cbrs.rank(test_ratings, methodology=TrainingItemsMethodology(), num_cpus=1)
        for user in set(test_ratings.user_id_column):
            result_single = result_different_meth.get_user_interactions(user)
            result_single_items = set([result_interaction.item_id for result_interaction in result_single])
            items_already_seen_user = set([train_interaction.item_id
                                           for train_interaction in train_ratings.get_user_interactions(user)])
            items_expected_rank = set([train_interaction.item_id
                                       for train_interaction in train_ratings
                                       if train_interaction.item_id not in items_already_seen_user])

            self.assertEqual(items_expected_rank, result_single_items)

        # Test algorithm not fitted
        cbrs = ContentBasedRS(alg, train_ratings_some_missing, self.movies_multiple)

        cbrs.fit(num_cpus=1)
        result_empty = cbrs.rank(test_ratings, user_id_list=['A000'], num_cpus=1)
        self.assertTrue(len(result_empty) == 0)

    def test_predict(self):
        # Test fit with the cbrs algorithm
        alg = LinearPredictor({'Plot': ['tfidf', 'embedding']}, SkLinearRegression())
        cbrs = ContentBasedRS(alg, train_ratings, self.movies_multiple)

        # we must fit the algorithm in order to predict
        cbrs.fit(num_cpus=1)

        # Test predict with the cbrs algorithm on specified items
        result_predict_filtered = cbrs.predict(test_ratings, num_cpus=1)
        self.assertEqual(len(result_predict_filtered), len(test_ratings))

        # Test predict with the cbrs algorithm on all available unseen items
        result_predict_all = cbrs.predict(test_ratings, methodology=None, num_cpus=1)
        self.assertTrue(len(result_predict_all) != 0)

        # Test predict with the cbrs algorithm for only some users
        result_predict_subset = cbrs.predict(test_ratings, methodology=None, user_id_list=["A000", "A003"], num_cpus=1)
        self.assertEqual(set(result_predict_subset.user_id_column), {"A000", "A003"})
        for user in {"A000", "A003"}:
            result_single = result_predict_subset.get_user_interactions(user)
            self.assertTrue(len(result_single) != 0)

        # Test predict with alternative methodology
        result_different_meth = cbrs.predict(test_ratings, methodology=TrainingItemsMethodology(), num_cpus=1)
        for user in set(test_ratings.user_id_column):
            result_single = result_different_meth.get_user_interactions(user)
            result_single_items = set([result_interaction.item_id for result_interaction in result_single])
            items_already_seen_user = set([train_interaction.item_id
                                           for train_interaction in train_ratings.get_user_interactions(user)])
            items_expected_rank = set([train_interaction.item_id
                                       for train_interaction in train_ratings
                                       if train_interaction.item_id not in items_already_seen_user])

            self.assertEqual(items_expected_rank, result_single_items)

        # Test algorithm not fitted
        cbrs = ContentBasedRS(alg, train_ratings_some_missing, self.movies_multiple)

        cbrs.fit(num_cpus=1)
        result_empty = cbrs.predict(test_ratings, user_id_list=['A000'], num_cpus=1)
        self.assertTrue(len(result_empty) == 0)

    def test_predict_raise_error(self):
        alg = ClassifierRecommender({'Plot': ['tfidf', 'embedding']}, SkSVC())
        cbrs = ContentBasedRS(alg, train_ratings, self.movies_multiple)

        # You must fit first in order to predict
        cbrs.fit(num_cpus=1)

        # This will raise error since page rank is not a prediction algorithm
        with self.assertRaises(NotPredictionAlg):
            cbrs.predict(test_ratings)

    def test_fit_rank(self):
        alg = LinearPredictor({'Plot': ['tfidf', 'embedding']}, SkLinearRegression())
        cbrs = ContentBasedRS(alg, train_ratings, self.movies_multiple)

        result = cbrs.fit_rank(test_ratings, save_fit=True, num_cpus=1)

        self.assertTrue(len(result) != 0)

        # with self.assertRaises(NotFittedAlg):
        #     cbrs.rank(test_ratings)


    def test_fit_predict(self):
        alg = LinearPredictor({'Plot': ['tfidf', 'embedding']}, SkLinearRegression())
        cbrs = ContentBasedRS(alg, train_ratings, self.movies_multiple)

        result = cbrs.fit_predict(test_ratings, num_cpus=1)

        self.assertTrue(len(result) != 0)


class TestGraphBasedRS(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        # different test ratings from the cbrs since a graph based algorithm
        # can give predictions only to items that are in the graph
        test_ratings = pd.DataFrame.from_records([
            ("A000", "tt0112896", None),
            ("A000", "tt0112453", None),
            ("A001", "tt0114576", None),
            ("A001", "tt0113497", None),
            ("A002", "tt0114576", None),
            ("A002", "tt0113041", None),
            ("A003", "tt0114576", None)],
            columns=["from_id", "to_id", "score"])
        cls.test_ratings = Ratings.from_dataframe(test_ratings)

        train_ratings = pd.DataFrame.from_records([
            ("A000", "tt0114576", 5, "54654675"),
            ("A001", "tt0114576", 3, "54654675"),
            ("A001", "tt0112896", 1, "54654675"),
            ("A000", "tt0113041", 1, "54654675"),
            ("A002", "tt0112453", 2, "54654675"),
            ("A002", "tt0113497", 4, "54654675"),
            ("A003", "tt0112453", 1, "54654675"),
            ("A003", "tt0113497", 4, "54654675")],
            columns=["from_id", "to_id", "score", "timestamp"])
        train_ratings = Ratings.from_dataframe(train_ratings)

        cls.graph = NXFullGraph(train_ratings)

    def test_rank(self):
        # Test rank with the graph based algorithm
        alg = NXPageRank()
        gbrs = GraphBasedRS(alg, self.graph)

        # Test ranking with the graph based algorithm on specified items
        result_rank_filtered = gbrs.rank(self.test_ratings, num_cpus=1)
        self.assertEqual(len(result_rank_filtered), len(self.test_ratings))

        # Test ranking with the gbrs algorithm on all unseen items that are in the graph
        result_rank_all = gbrs.rank(self.test_ratings, methodology=None, num_cpus=1)
        self.assertTrue(len(result_rank_all) != 0)

        # Test top-n ranking with the gbrs algorithm only for some users
        result_rank_numbered = gbrs.rank(self.test_ratings, n_recs=2, methodology=None, user_id_list=["A000", "A003"],
                                         num_cpus=1)
        self.assertEqual(set(result_rank_numbered.user_id_column), {"A000", "A003"})
        for user in {"A000", "A003"}:
            result_single = result_rank_numbered.get_user_interactions(user)
            self.assertTrue(len(result_single) == 2)

        # Test ranking with alternative methodology
        result_different_meth = gbrs.rank(self.test_ratings, methodology=TrainingItemsMethodology(), num_cpus=1)
        for user in set(self.test_ratings.user_id_column):
            result_single = set([pred_rank.item_id for pred_rank in result_different_meth if pred_rank.user_id == user])
            items_already_seen_user = set([original_interaction.item_id
                                           for original_interaction in train_ratings.get_user_interactions(user)])
            items_expected_rank = set([train_interaction.item_id
                                       for train_interaction in train_ratings
                                       if train_interaction.item_id not in items_already_seen_user])

            self.assertEqual(items_expected_rank, result_single)

    def test_predict_raise_error(self):
        alg = NXPageRank()
        gbrs = GraphBasedRS(alg, self.graph)

        # This will raise error since page rank is not a prediction algorithm
        with self.assertRaises(NotPredictionAlg):
            gbrs.predict(self.test_ratings, num_cpus=1)

    def test_rank_filterlist_empty_A000(self):
        # no items to recommend is in the graph for user A000
        test_ratings = pd.DataFrame.from_records([
            ("A000", "not_in_graph", None),
            ("A000", "not_in_graph1", None),
            ("A001", "tt0114576", None),
            ("A001", "tt0113497", None),
            ("A002", "tt0114576", None),
            ("A002", "tt0113041", None),
            ("A003", "tt0114576", None)],
            columns=["from_id", "to_id", "score"])
        test_ratings = Ratings.from_dataframe(test_ratings)

        # Test rank with the graph based algorithm
        alg = NXPageRank()
        gbrs = GraphBasedRS(alg, self.graph)

        # Test ranking with the graph based algorithm with items not present in the graph for A000
        result_rank = gbrs.rank(test_ratings, num_cpus=1)
        self.assertTrue(len(result_rank) != 0)

        # no rank is present for A000
        with self.assertRaises(KeyError):
            result_rank.get_user_interactions('A000')

    def test_rank_filterlist_empty_all(self):
        # different test ratings from the cbrs since a graph based algorithm
        # can give predictions only to items that are in the graph
        test_ratings = pd.DataFrame.from_records([
            ("A000", "not_in_graph", None),
            ("A000", "not_in_graph1", None),
            ("A001", "not_in_graph2", None),
            ("A001", "not_in_graph3", None),
            ("A002", "not_in_graph4", None),
            ("A002", "not_in_graph5", None),
            ("A003", "not_in_graph6", None)],
            columns=["from_id", "to_id", "score"])
        test_ratings = Ratings.from_dataframe(test_ratings)

        # Test rank with the graph based algorithm
        alg = NXPageRank()
        gbrs = GraphBasedRS(alg, self.graph)

        # Test ranking with the graph based algorithm on items not in the graph, we expect it to be empty
        result_rank_empty = gbrs.rank(test_ratings, num_cpus=1)
        self.assertTrue(len(result_rank_empty) == 0)



import csv
import os
from abc import ABC, abstractmethod

import json
from typing import Dict, Iterator

import mysql.connector


class RawInformationSource(ABC):
    """
    Abstract Class that generalizes the acquisition of raw descriptions of the contents
    from one of the possible raw sources.

    Args:
        encoding: define the type of encoding of data stored in the source (example: "utf-8")
    """

    def __init__(self, encoding: str):
        self.__encoding = encoding

    @property
    def encoding(self):
        return self.__encoding

    @property
    @abstractmethod
    def representative_name(self):
        raise NotImplementedError

    @abstractmethod
    def __iter__(self) -> Iterator[Dict[str, str]]:
        """
        Iter on contents in the source, each iteration returns a dict representing a "row" in the raw content
        """
        raise NotImplementedError

    @abstractmethod
    def __len__(self):
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class DATFile(RawInformationSource):
    """
    Wrapper for a DAT file. This class is able to read from a DAT file where each entry is separated by the `::` string.
    Since a DAT file has no header, each entry can be referenced with a string representing its positional index
    (e.g. '0' for entry in the first position, '1' for the entry in the second position, etc.)

    You can iterate over the whole content of the raw source with a simple for loop: each row will be returned as a
    dictionary where keys are strings representing the positional indices, values are the entries

    Examples:
        Consider the following DAT file
        ```
        10::worker::75011
        11::without occupation::76112
        ```

        >>> file = DATFile(dat_path)
        >>> print(list(file))
        [{'0': '10', '1': 'worker', '2': '75011'},
        {'0': '11', '1': 'without occupation', '2': '76112'}]

    Args:
        file_path: path of the dat file
        encoding: define the type of encoding of data stored in the source (example: "utf-8")
    """

    def __init__(self, file_path: str, encoding: str = "utf-8"):
        super().__init__(encoding)
        self.__file_path = file_path

    @property
    def representative_name(self) -> str:
        """
        Method which returns a meaningful name for the raw source.

        In this case it's simply the file name + its extension

        Returns:
            The representative name for the raw source
        """
        # file name with extension
        file_name = os.path.basename(self.__file_path)

        return file_name

    def __iter__(self) -> Iterator[Dict[str, str]]:
        with open(self.__file_path, encoding=self.encoding) as f:
            for line in f:
                line_dict = {}
                fields = line.split('::')
                for i, field in enumerate(fields):
                    field = field.strip("\n\t\r")
                    line_dict[str(i)] = field

                yield line_dict

    def __len__(self):
        with open(self.__file_path, newline='', encoding=self.encoding) as dat_file:
            total_length = sum(1 for _ in dat_file)

            return total_length

    def __str__(self):
        return "DATFile"

    def __repr__(self):
        return f'DATFile(encoding={self.__encoding}, file_path={self.__file_path})'


class JSONFile(RawInformationSource):
    """
    Wrapper for a JSON file. This class is able to read from a JSON file where each "row" is a dictionary-like object
    inside a list

    You can iterate over the whole content of the raw source with a simple for loop: each row will be returned as a
    dictionary

    Examples:
        Consider the following JSON file
        ```
        [{"Title":"Jumanji","Year":"1995"},
         {"Title":"Toy Story","Year":"1995"}]
        ```

        >>> file = JSONFile(json_path)
        >>> print(list(file))
        [{'Title': 'Jumanji', 'Year': '1995'},
         {'Title': 'Toy Story', 'Year': '1995'}]

    Args:
        file_path: path of the dat file
        encoding: define the type of encoding of data stored in the source (example: "utf-8")
    """

    def __init__(self, file_path: str, encoding: str = "utf-8"):
        super().__init__(encoding)
        self.__file_path = file_path

    @property
    def representative_name(self) -> str:
        """
        Method which returns a meaningful name for the raw source.

        In this case it's simply the file name + its extension

        Returns:
            The representative name for the raw source
        """
        # file name with extension
        file_name = os.path.basename(self.__file_path)

        return file_name

    def __iter__(self) -> Iterator[Dict[str, str]]:
        with open(self.__file_path, encoding=self.encoding) as j:
            all_lines = json.load(j, parse_int=str, parse_float=str)
            for line in all_lines:
                yield line

    def __len__(self):
        with open(self.__file_path, encoding=self.encoding) as j:
            return len(json.load(j))

    def __str__(self):
        return "JSONFile"

    def __repr__(self):
        return f'JSONFile(encoding={self.__encoding}, file_path={self.__file_path})'


class CSVFile(RawInformationSource):
    r"""
    Wrapper for a CSV file. This class is able to read from a CSV file where each entry is separated by the a certain
    separator (`,` by default). So by using this class you can also read TSV file for examples, by specifying
    `separator='\t'`.

    A CSV File most typically has a header: in this case, each entry can be referenced with its column header.
    In case the CSV File hasn't a header, simply specify `has_header=False`: in this case, each entry can be referenced
    with a string representing its positional index
    (e.g. '0' for entry in the first position, '1' for the entry in the second position, etc.)

    You can iterate over the whole content of the raw source with a simple for loop: each row will be returned as a
    dictionary where keys are strings representing the positional indices, values are the entries

    Examples:

        Consider the following CSV file with header
        ```
        movie_id,movie_title,release_year
        1,Jumanji,1995
        2,Toy Story,1995
        ```

        >>> file = CSVFile(csv_path)
        >>> print(list(file))
        [{'movie_id': '1', 'movie_title': 'Jumanji', 'release_year': '1995'},
        {'movie_id': '2', 'movie_title': 'Toy Story', 'release_year': '1995'}]

        Consider the following TSV file with no header
        ```
        1   Jumanji 1995
        2   Toy Story   1995
        ```

        >>> file = CSVFile(tsv_path, separator='\t', has_header=False)
        >>> print(list(file))
        [{'0': '1', '1': 'Jumanji', '2': '1995'},
        {'0': '2', '1': 'Toy Story', '2': '1995'}]

    Args:
        file_path: Path of the dat file
        separator: Character which separates each entry. By default is a comma (`,`), but in case you need to read from
            a TSV file simply change this parameter to `\t`
        has_header: Boolean value which specifies if the file has an header or not. Default is True
        encoding: Define the type of encoding of data stored in the source (example: "utf-8")
    """

    def __init__(self, file_path: str, separator: str = ',', has_header: bool = True, encoding: str = "utf-8-sig"):
        super().__init__(encoding)
        self.__file_path = file_path
        self.__has_header = has_header
        self.__separator = separator

    @property
    def representative_name(self) -> str:
        """
        Method which returns a meaningful name for the raw source.

        In this case it's simply the file name + its extension

        Returns:
            The representative name for the raw source
        """
        # file name with extension
        file_name = os.path.basename(self.__file_path)

        return file_name

    def __iter__(self) -> Iterator[Dict[str, str]]:
        with open(self.__file_path, newline='', encoding=self.encoding) as csv_file:
            if self.__has_header:
                reader = csv.DictReader(csv_file, quoting=csv.QUOTE_MINIMAL, delimiter=self.__separator)
            else:
                reader = csv.DictReader(csv_file, quoting=csv.QUOTE_MINIMAL, delimiter=self.__separator)
                reader.fieldnames = [str(i) for i in range(len(reader.fieldnames))]
                csv_file.seek(0)

            yield from reader

    def __len__(self):
        with open(self.__file_path, newline='', encoding=self.encoding) as csv_file:
            total_length = sum(1 for _ in csv_file)
            if self.__has_header:
                total_length -= 1

            return total_length

    def __str__(self):
        return "CSVFile"

    def __repr__(self):
        return f'CSVFile(file_path={self.__file_path}, separator={self.__separator}, has_header={self.__has_header}, ' \
               f'encoding={self.encoding})'


class SQLDatabase(RawInformationSource):
    """
    Wrapper for a SQL database.

    You can iterate over the whole content of the raw source with a simple for loop: each row will be returned as a
    dictionary where keys are strings representing the positional indices, values are the entries

    Examples:

        Consider the following SQL table for the databaase 'movies' in localhost
        ```
        +----------+-------------+--------------+
        | Movie ID | Movie Title | Release Year |
        +----------+-------------+--------------+
        |        1 | Jumanji     |         1995 |
        |        2 | Toy Story   |         1995 |
        +----------+-------------+--------------+
        ```

        >>> file = SQLDatabase(host='127.0.0.1', username='root', password='root',
        >>>                    database_name='movies', table_name='movies_table')
        >>> print(list(file))
        [{'Movie ID': '1', 'Movie Title': 'Jumanji', 'Release Year': '1995'},
        {'Movie ID': '2', 'Movie Title': 'Toy Story', 'Release Year': '1995'}]

    Args:
        host: host ip of the sql server
        username: username for the access
        password: password for the access
        database_name: name of database
        table_name: name of the database table where data is stored
        encoding: Define the type of encoding of data stored in the source (example: "utf-8")
    """

    def __init__(self, host: str,
                 username: str,
                 password: str,
                 database_name: str,
                 table_name: str,
                 encoding: str = "utf-8"):
        super().__init__(encoding)
        self.__host: str = host
        self.__username: str = username
        self.__password: str = password
        self.__database_name: str = database_name
        self.__table_name: str = table_name

        conn = mysql.connector.connect(host=self.__host,
                                       user=self.__username,
                                       password=self.__password,
                                       charset=self.encoding)
        cursor = conn.cursor()
        query = """USE """ + self.__database_name + """;"""
        cursor.execute(query)
        conn.commit()
        self.__conn = conn

    @property
    def host(self) -> str:
        return self.__host

    @property
    def username(self) -> str:
        return self.__username

    @property
    def password(self) -> str:
        return self.__password

    @property
    def database_name(self) -> str:
        return self.__database_name

    @property
    def table_name(self) -> str:
        return self.__table_name

    @property
    def conn(self):
        return self.__conn

    @host.setter
    def host(self, host: str):
        self.__host = host

    @username.setter
    def username(self, username: str):
        self.__username = username

    @password.setter
    def password(self, password: str):
        self.__password = password

    @database_name.setter
    def database_name(self, database_name: str):
        self.__database_name = database_name

    @table_name.setter
    def table_name(self, table_name: str):
        self.__table_name = table_name

    @conn.setter
    def conn(self, conn):
        self.__conn = conn

    @property
    def representative_name(self) -> str:
        """
        Method which returns a meaningful name for the raw source.

        In this case it's the host name followed by the table name

        Returns:
            The representative name for the raw source
        """
        return f"{self.host}/{self.table_name}"

    def __iter__(self) -> Iterator[Dict[str, str]]:
        cursor = self.conn.cursor(dictionary=True)
        query = """SELECT * FROM """ + self.table_name + """;"""
        cursor.execute(query)
        for result in cursor:
            yield result

    def __len__(self):
        cursor = self.conn.cursor()
        query = """SELECT * FROM """ + self.table_name + """;"""
        cursor.execute(query)
        cursor.fetchall()

        return cursor.rowcount

    def __str__(self):
        return "SQLDatabase"

    def __repr__(self):
        return f'SQLDatabase(host={self.__host},' \
               f'username={self.__username}, password={self.__password}, ' \
               f'database_name={self.__database_name}, table_name={self.__table_name}, encoding={self.encoding})'



import inspect


def autorepr(obj, frame):
    # pull tuple from frame
    args, args_paramname, kwargs_paramname, values = inspect.getargvalues(frame)

    args = args[1:]  # remove 'self' argument from function

    arg_string = ''

    # add to arg string formal argument
    arg_string += ', '.join([f"{arg}={repr(values[arg])}"
                             for arg in (args if args is not None else [])])

    # show positional varargs
    if args_paramname is not None:
        varglist = values[args_paramname]
        if len(arg_string) != 0:
            arg_string += ', '
        arg_string += ', '.join([f"*{args_paramname}={repr(v)}"
                                 for v in (varglist if varglist is not None else [])])

    # show named varargs
    if kwargs_paramname is not None:
        varglist = values[kwargs_paramname]
        if len(arg_string) != 0:
            arg_string += ', '
        arg_string += ', '.join([f"*{kwargs_paramname}_{k}={repr(varglist[k])}"
                                 for k in (sorted(varglist) if varglist is not None else [])])

    name_obj = obj.__class__.__name__
    repr_string = f"{name_obj}({arg_string})"

    return repr_string



from random import random
from unittest import mock
from unittest.mock import patch, Mock, MagicMock
import numpy as np

from test.content_analyzer.embeddings.test_embedding_source import TestEmbeddingSource
from clayrs.content_analyzer.embeddings.embedding_loader.gensim import Gensim

result_matrix = {
    'title': np.array([random() for _ in range(25)]),
    'plot': np.array([random() for _ in range(25)])
}


def get_item(key):
    return result_matrix[key]


def similar_by_vector(vector, n_to_find):
    for i, vec in enumerate(result_matrix.values()):
        if np.array_equal(vec, vector):
            return [(list(result_matrix.keys())[i], 1)]


mocked_model = MagicMock()
mocked_model.__getitem__.side_effect = get_item
#mocked_model = MagicMock()
mocked_model.similar_by_vector.side_effect = similar_by_vector
mocked_model.vector_size = 25


class TestGensimDownloader(TestEmbeddingSource):

    def test_load(self):

        with mock.patch('gensim.downloader.info', return_value={'models': 'glove-twitter-25'}):
            with mock.patch('gensim.downloader.load', return_value=mocked_model):
                source = Gensim('glove-twitter-25')

        # result is a matrix containing 2 rows, one for 'title', one for 'plot'
        result = source.load(["title", "plot"])

        # the expected shape of result is (2, 25):
        # 2 for words and 25 due to the model 'glove-twitter-25'
        expected_shape = (2, 25)
        self.assertEqual(expected_shape, result.shape)

        self.assertWordEmbeddingMatches(source, result[0], "title")
        self.assertWordEmbeddingMatches(source, result[1], "plot")
